import pathlib
import uuid
from typing import Dict, List

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class AddColumnsTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvAddColumnsAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.ADD_COLUMN:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvAddColumnsAction = action

        column_data: Dict[int, actions.TsvColumnData] = (
            action.columns if action.columns else {}
        )
        if not column_data:
            result.message = "There is not column index"
            return result

        cell_default_values: Dict[int, str] = (
            action.cell_default_values if action.cell_default_values else {}
        )

        column_indices: List[int] = list(column_data.keys()).copy()
        column_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with source_file_path.open("r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")
                for column_idx in column_indices:
                    if column_idx < 0:
                        column_idx = len(header_names)
                    default_value = (
                        cell_default_values[column_idx]
                        if column_idx in cell_default_values
                        and cell_default_values[column_idx]
                        else ""
                    )
                    value: actions.TsvColumnData = column_data[column_idx]
                    if not value or not value.header_name:
                        result.message = (
                            f"There is not header name for column index {column_idx}"
                        )
                        return result
                    header_names.insert(column_idx, value.header_name)

                with target_file_path.open("w", encoding=write_encoding) as target:
                    self.write_row(target, header_names)
                    row_index = 0
                    for line in source:
                        row = line.strip("\n").split("\t")
                        for column_idx in column_indices:
                            default_value = (
                                cell_default_values[column_idx]
                                if column_idx in cell_default_values
                                and cell_default_values[column_idx]
                                else ""
                            )
                            coloumn_data: actions.TsvColumnData = column_data[
                                column_idx
                            ]
                            value = (
                                coloumn_data.values[row_index]
                                if row_index in coloumn_data.values
                                else default_value
                            )
                            row.insert(column_idx, value)
                        self.write_row(target, row)
                        row_index += 1
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



from typing import List

from metabolights_utils.isatab.default.base_isa_table_file import BaseIsaTableFileReader


class DefaultAssignmentFileReader(BaseIsaTableFileReader):
    patterns = [
        ["^(datdatabase_identifier)$", ""],
        ["^(chemical_formula)$", ""],
        ["^(smiles)$", ""],
        ["^(inchi)$", ""],
        ["^(metabolite_identification)$", ""],
        ["^(mass_to_charge)$", ""],
        ["^(fragmentation)$", ""],
        ["^(modifications)$", ""],
        ["^(charge)$", ""],
        ["^(retention_time)$", ""],
        ["^(taxid)$", ""],
        ["^(chemical_shift)$", ""],
        ["^(multiplicity)$", ""],
        ["^(species)$", ""],
        ["^(database)$", ""],
        ["^(database_version)$", ""],
        ["^(reliability)$", ""],
        ["^(uri)$", ""],
        ["^(search_engine)$", ""],
        ["^(search_engine_score)$", ""],
        ["^(smallmolecule_abundance_sub)$", ""],
        ["^(smallmolecule_abundance_stdev_sub)$", ""],
        ["^(smallmolecule_abundance_std_error_sub)$", ""],
    ]

    def __init__(self, results_per_page=100) -> None:
        super().__init__(results_per_page=results_per_page)

    def get_expected_patterns(self) -> List[List[str]]:
        return DefaultAssignmentFileReader.patterns



from pathlib import Path
from typing import Union

import click

from metabolights_utils.provider import definitions
from metabolights_utils.provider.ftp.model import LocalDirectory
from metabolights_utils.provider.ftp_repository import MetabolightsFtpRepository
from metabolights_utils.provider.utils import is_metadata_filename_pattern


@click.command(name="download")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_repository_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--ftp_server_url",
    "-f",
    default=f"{definitions.default_ftp_server_url}",
    help="FTP server URL where MetaboLights repository is hosted. ",
)
@click.option(
    "--ftp_root_directory",
    "-d",
    default=f"{definitions.default_remote_repository_root_directory}",
    help="MetaboLights study directory on FTP server URL. ",
)
@click.option(
    "--local_cache_path",
    "-c",
    default=definitions.default_local_repority_cache_path,
    help="Path to store cache files of FTP file indices, study models, etc.",
)
@click.option(
    "--override_local_files",
    "-o",
    is_flag=True,
    default=False,
    help="Downloads files and override current local copies.",
)
@click.argument("study_id")
@click.argument("file", required=False)
def public_download(
    study_id: Union[None, str] = None,
    file: Union[None, str] = None,
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    ftp_server_url: Union[None, str] = None,
    ftp_root_directory: Union[None, str] = None,
    override_local_files: bool = False,
):
    """
    Download study data and metadata files from MetaboLights FTP server.

    study_id: MetaboLights study accession number (MTBLSxxxx).

    file (optional): Relative file path in study folder. All ISA metadata files will be downloaded if not specified.
    """
    study_id = study_id.upper()
    client = MetabolightsFtpRepository(
        ftp_server_url=ftp_server_url,
        remote_repository_root_directory=ftp_root_directory,
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
    )
    is_metadata_file = is_metadata_filename_pattern(file)
    local_path = client.local_storage_root_path
    if not file or is_metadata_file:
        metadata_files = [file] if file else None
        result: LocalDirectory = client.download_study_metadata_files(
            study_id=study_id,
            local_path=local_path,
            metadata_files=metadata_files,
            override_local_files=override_local_files,
            delete_unlisted_metadata_files=False,
        )
    else:
        result: LocalDirectory = client.download_study_data_files(
            study_id=study_id,
            local_path=local_path,
            selected_data_files=[file] if file else None,
            override_local_files=override_local_files,
            skip_files=None,
            delete_unlisted_local_files=False,
            keep_local_files=None,
        )

    if result.success:
        if result.actions:
            local_study_path = Path(local_path) / Path(study_id)
            click.echo(
                click.style(f"Downloaded files on {local_study_path}", fg="green")
            )
            for item, action in result.actions.items():
                click.echo(f"{action}\t{item.replace(f'{study_id}/', '', 1)}")
        else:
            click.echo("There is no ISA metadata file to download.")
    else:
        click.echo(f"FTP response error: {result.message}")
        exit(1)



import datetime
import logging
import shutil
from pathlib import Path
from typing import Union

from metabolights_utils.utils.filename_utils import join_path
from metabolights_utils.utils.search_utils import MetabolightsSearchUtils as SearchUtils

logger = logging.getLogger(__name__)


class MetabolightsAuditUtils:
    @staticmethod
    def create_audit_folder(
        src_root_path: str,
        target_root_path: str,
        folder_suffix: Union[None, str] = "BACKUP",
        folder_prefix: Union[None, str] = None,
        timestamp_format: str = "%Y-%m-%d_%H-%M-%S",
    ) -> Union[None, str]:
        if not timestamp_format or not src_root_path or not target_root_path:
            logger.error("Invalid input parameter.")
            return None
        base = datetime.datetime.now(datetime.timezone.utc).strftime(timestamp_format)
        folder_name = f"{base}_{folder_suffix}" if folder_suffix else base
        folder_name = f"{folder_prefix}_{folder_name}" if folder_prefix else folder_name
        logger.info("Audit folder name: %s", folder_name)
        target_folder_path = join_path(target_root_path, folder_name)
        return MetabolightsAuditUtils.copy_isa_metadata_files(
            src_folder_path=src_root_path,
            target_folder_path=target_folder_path,
        )

    @staticmethod
    def copy_isa_metadata_files(
        src_folder_path: str, target_folder_path: str
    ) -> Union[None, str]:
        source = Path(src_folder_path)
        target = Path(target_folder_path)
        real_src_folder_path = source.resolve()
        real_target_folder_path = target.resolve()
        metadata_files_list = SearchUtils.get_isa_metadata_files(
            folder_path=real_src_folder_path, recursive=False
        )
        metadata_files_list.sort()

        if metadata_files_list:
            if not real_target_folder_path.exists():
                real_target_folder_path.mkdir(parents=True, exist_ok=True)
                logger.info("Target folder is created: %s", target_folder_path)
            for file in metadata_files_list:
                basename = Path(file).name
                target_file = join_path(target_folder_path, basename)
                shutil.copy2(file, target_file, follow_symlinks=False)
            logger.info(
                "Metadata files are copied from %s to %s",
                src_folder_path,
                target_folder_path,
            )
            return target_folder_path
        logger.warning("There is no metadata file on %s", src_folder_path)
        return None



from typing import List, Literal, OrderedDict, Union

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.isa.common import Comment, IsaAbstractModel, IsaTabConfig

module_name = __name__


class BaseSection(IsaAbstractModel):
    comments: Annotated[
        List[Comment],
        Field(
            description="List of comments linked to the related section",
        ),
    ] = []


class OntologySourceReference(IsaAbstractModel):
    isatab_config: Annotated[IsaTabConfig, Field(exclude=True)] = IsaTabConfig(
        field_order=[
            "source_name",
            "source_file",
            "source_version",
            "source_description",
        ]
    )

    source_name: Annotated[
        str,
        Field(
            description="The name of the source of a term; "
            "i.e. the source controlled vocabulary or ontology."
            " These names will be used in all corresponding Term Source REF fields "
            "that occur elsewhere.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Source Name",
                "source_description": True,
            },
        ),
    ] = ""

    source_file: Annotated[
        str,
        Field(
            description="A file name or a URI of an official resource.",
            json_schema_extra={"auto_fill": True, "header_name": "Source File"},
        ),
    ] = ""
    source_version: Annotated[
        Union[str, int],
        Field(
            description="The version number of the Term Source "
            "to support terms tracking.",
            json_schema_extra={"auto_fill": True, "header_name": "Source Version"},
        ),
    ] = ""
    source_description: Annotated[
        str,
        Field(
            description="Description of source. "
            "Use for disambiguating resources when homologous prefixes have been used.",
            json_schema_extra={"auto_fill": True, "header_name": "Source Description"},
        ),
    ] = ""


class OntologyAnnotation(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(exclude=True),
    ] = IsaTabConfig(field_order=["term", "term_accession_number", "term_source_ref"])

    term: Annotated[
        str,
        Field(
            description="Ontology term",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
            },
        ),
    ] = ""

    term_accession_number: Annotated[
        str,
        Field(
            description="The accession number from the Term Source "
            "associated with the term.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Term Accession Number",
            },
        ),
    ] = ""
    term_source_ref: Annotated[
        str,
        Field(
            description="Source reference name of ontology term. e.g., EFO, OBO, NCIT. "
            "Ontology source reference names should be defined "
            "in ontology source references section in i_Investigation.txt file",
            json_schema_extra={"auto_fill": True, "header_name": "Term Source REF"},
        ),
    ] = ""

    def __str__(self) -> str:
        if self.term:
            if self.term_source_ref or self.term_accession_number:
                return (
                    f"{self.term} [{self.term_source_ref}:{self.term_accession_number}]"
                )
            else:
                return self.term
        return ""


class ExtendedOntologyAnnotation(OntologyAnnotation):
    category: Annotated[
        str,
        Field(description="Category of descriptor annotation"),
    ] = ""
    source: Annotated[
        str,
        Field(description="Source of ontology annotation"),
    ] = ""


class ValueTypeAnnotation(IsaAbstractModel):
    isatab_config: Annotated[IsaTabConfig, Field(exclude=True)] = IsaTabConfig(
        field_order=[
            "name",
            "type",
            "term_accession_number",
            "term_source_ref",
        ]
    )

    name: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Name"}
        ),
    ] = ""
    type: Annotated[
        str,
        Field(
            description="", json_schema_extra={"auto_fill": True, "header_name": "Type"}
        ),
    ] = ""
    term_accession_number: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Type Term Accession Number",
            },
        ),
    ] = ""
    term_source_ref: Annotated[
        str,
        Field(
            description="",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Type Term Source REF",
            },
        ),
    ] = ""


class CharacteristicDefinition(ValueTypeAnnotation):
    value_format: Annotated[
        Literal["", "text", "ontology", "numeric"],
        Field(
            description="Format of characteristic value, "
            "empty value is used for undefined value."
            "text (one column), "
            "ontology ( one column+ term source ref and accession number), "
            "numeric (one column + unit, unit term source ref and accession number)",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""


class ParameterDefinition(OntologyAnnotation):
    value_format: Annotated[
        Literal["", "text", "ontology", "numeric"],
        Field(
            description="Format of parameter value, "
            "empty value is used for undefined value."
            "text (one column), "
            "ontology ( one column+ term source ref and accession number), "
            "numeric (one column + unit, unit term source ref and accession number)",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""


class ProtocolComponent(ValueTypeAnnotation): ...


class Publication(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["pub_med_id", "doi", "author_list", "title", "status"]
    )

    pub_med_id: Annotated[
        str,
        Field(
            description="The PubMed IDs of the publication.",
            json_schema_extra={"auto_fill": True, "header_name": "PubMed ID"},
            title="PubMed Id",
        ),
    ] = ""
    doi: Annotated[
        str,
        Field(
            description="A Digital Object Identifier (DOI) for the publication.",
            json_schema_extra={"auto_fill": True, "header_name": "Publication DOI"},
            title="DOI",
        ),
    ] = ""
    author_list: Annotated[
        str,
        Field(
            description="The list of authors associated with the publication. "
            "Comma (,) is recommended to define multiple authors.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Publication Author List",
            },
        ),
    ] = ""
    title: Annotated[
        str,
        Field(
            description="The title of publication associated with the investigation.",
            json_schema_extra={"auto_fill": True, "header_name": "Publication Title"},
        ),
    ] = ""
    status: Annotated[
        OntologyAnnotation,
        Field(
            description="A term describing the status of that publication "
            "(i.e. EFO:submitted, EFO:in preparation, EFO:published).",
            json_schema_extra={"auto_fill": True, "header_name": "Publication Status"},
        ),
    ] = OntologyAnnotation()


class Person(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "last_name",
            "first_name",
            "mid_initials",
            "email",
            "phone",
            "fax",
            "address",
            "affiliation",
            "roles",
        ]
    )

    last_name: Annotated[
        str,
        Field(
            description="Last name of a person "
            "associated with the investigation or study.",
            json_schema_extra={"auto_fill": True, "header_name": "Last Name"},
        ),
    ] = ""
    first_name: Annotated[
        str,
        Field(
            description="First name of person "
            "associated with the investigation or study.",
            json_schema_extra={"auto_fill": True, "header_name": "First Name"},
        ),
    ] = ""
    mid_initials: Annotated[
        str,
        Field(
            description="Middle name initials (if exists) of person "
            "associated with the investigation or study",
            json_schema_extra={"auto_fill": True, "header_name": "Mid Initials"},
        ),
    ] = ""
    email: Annotated[
        str,
        Field(
            description="Email address of person "
            "associated with the investigation or study",
            json_schema_extra={"auto_fill": True, "header_name": "Email"},
        ),
    ] = ""
    phone: Annotated[
        str,
        Field(
            description="Phone number of person "
            "associated with the investigation or study",
            json_schema_extra={"auto_fill": True, "header_name": "Phone"},
        ),
    ] = ""
    fax: Annotated[
        str,
        Field(
            description="Fax number of person "
            "associated with the investigation or study",
            json_schema_extra={"auto_fill": True, "header_name": "Fax"},
        ),
    ] = ""
    address: Annotated[
        str,
        Field(
            description="Address of person associated with the investigation or study",
            json_schema_extra={"auto_fill": True, "header_name": "Address"},
        ),
    ] = ""
    affiliation: Annotated[
        str,
        Field(
            description="Affiliation of person associated "
            "with the investigation or study",
            json_schema_extra={"auto_fill": True, "header_name": "Affiliation"},
        ),
    ] = ""
    roles: Annotated[
        List[OntologyAnnotation],
        Field(
            description="Roles of person associated with the investigation or study. "
            "Multiple role can be defined for each person. "
            "Role is defined as an ontology term. "
            "e.g., NCIT:Investigator, NCIT:Author",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Roles",
                "text_multiple_value": True,
                "seperator": ";",
            },
        ),
    ] = []
    orcid: Annotated[
        str,
        Field(
            description="ORCID of person associated with the investigation or study",
            json_schema_extra={"auto_fill": False},
            title="ORCID",
        ),
    ] = ""
    additional_emails: Annotated[
        list[str],
        Field(
            description="Additional email addresses of person associated "
            "with the investigation or study",
            json_schema_extra={"auto_fill": False},
        ),
    ] = []
    affiliation_ror_id: Annotated[
        str,
        Field(
            description="ROR Id of the person's affiliation",
            json_schema_extra={"auto_fill": False},
            title="Affiliation ROR Id",
        ),
    ] = ""


class Factor(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(field_order=["name", "type"])

    name: Annotated[
        str,
        Field(
            description="The name of one factor used in the Study files. "
            "A factor corresponds to an independent variable "
            "manipulated by the experimentalist "
            "with the intention to affect biological systems in a way "
            "that can be measured by an assay.",
            json_schema_extra={"auto_fill": True, "header_name": "Name"},
        ),
    ] = ""
    type: Annotated[
        OntologyAnnotation,
        Field(
            description="A term allowing the classification of the factor "
            "into categories. "
            "The term is a controlled vocabulary or an ontology",
            json_schema_extra={"auto_fill": True, "header_name": "Type"},
        ),
    ] = OntologyAnnotation()

    value_format: Annotated[
        Literal["", "text", "ontology", "numeric"],
        Field(
            description="Format of factor value, "
            "empty value is used for undefined value."
            "text (one column), "
            "ontology ( one column+ term source ref and accession number), "
            "numeric (one column + unit, unit term source ref and accession number)",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""

    def __str__(self):
        return f"{self.name if self.name else ''} ({str(self.type)})"


class Assay(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "file_name",
            "measurement_type",
            "technology_type",
            "technology_platform",
        ]
    )

    file_name: Annotated[
        str,
        Field(
            description="Assay file name. Expected filename pattern is a_*.txt",
            json_schema_extra={"auto_fill": True, "header_name": "File Name"},
        ),
    ] = ""
    measurement_type: Annotated[
        OntologyAnnotation,
        Field(
            description="A term to qualify what is being measured "
            "(e.g. metabolite identification).",
            json_schema_extra={"auto_fill": True, "header_name": "Measurement Type"},
        ),
    ] = OntologyAnnotation()
    technology_type: Annotated[
        OntologyAnnotation,
        Field(
            description="Term to identify the technology "
            "used to perform the measurement, "
            "e.g. NMR spectrometry assay, mass spectrometry assay",
            json_schema_extra={"auto_fill": True, "header_name": "Technology Type"},
        ),
    ] = OntologyAnnotation()
    technology_platform: Annotated[
        str,
        Field(
            description="Platform information such as assay technique name, "
            "polarity, column model, manufacturer, platform name.",
            json_schema_extra={"auto_fill": True, "header_name": "Technology Platform"},
        ),
    ] = ""

    assay_identifier: Annotated[
        str,
        Field(description="Assay identifier.", json_schema_extra={"auto_fill": False}),
    ] = ""
    assay_type: Annotated[
        OntologyAnnotation,
        Field(description="Assay type.", json_schema_extra={"auto_fill": False}),
    ] = OntologyAnnotation()
    omics_type: Annotated[
        OntologyAnnotation,
        Field(description="Assay type.", json_schema_extra={"auto_fill": False}),
    ] = OntologyAnnotation()
    assay_descriptors: Annotated[
        list[ExtendedOntologyAnnotation],
        Field(description="Assay descriptors.", json_schema_extra={"auto_fill": False}),
    ] = []
    result_file_format: Annotated[
        str,
        Field(
            description="result file format. e.g., MAF v2.0, mzTab-M v2.1, etc.",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""


class Protocol(IsaAbstractModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "name",
            "protocol_type",
            "description",
            "uri",
            "version",
            "parameters",
            "components",
        ]
    )

    name: Annotated[
        str,
        Field(
            description="Protocol name.",
            json_schema_extra={"auto_fill": True, "header_name": "Name"},
        ),
    ] = ""
    protocol_type: Annotated[
        OntologyAnnotation,
        Field(
            description="Term to classify the protocol.",
            json_schema_extra={"auto_fill": True, "header_name": "Type"},
        ),
    ] = OntologyAnnotation()
    description: Annotated[
        str,
        Field(
            description="Protocol description.",
            json_schema_extra={"auto_fill": True, "header_name": "Description"},
        ),
    ] = ""
    uri: Annotated[
        str,
        Field(
            description="Pointer to external protocol resources "
            "that can be accessed by their Uniform Resource Identifier (URI).",
            json_schema_extra={"auto_fill": True, "header_name": "URI"},
        ),
    ] = ""
    version: Annotated[
        str,
        Field(
            description="An identifier for the version to ensure protocol tracking..",
            json_schema_extra={"auto_fill": True, "header_name": "Version"},
        ),
    ] = ""
    parameters: Annotated[
        List[ParameterDefinition],
        Field(
            description="Protocol's parameters.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Parameters Name",
                "text_multiple_value": True,
                "seperator": ";",
            },
        ),
    ] = []
    components: Annotated[
        List[ProtocolComponent],
        Field(
            description="Protocol’s components; e.g. instrument names, "
            "software names, and reagents names.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "Components",
                "text_multiple_value": True,
                "seperator": ";",
            },
        ),
    ] = []


######################################################################################################
# ISA TAB FILE SECTIONS
######################################################################################################


class OntologySourceReferences(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["references"],
        section_header="ONTOLOGY SOURCE REFERENCE",
        section_prefix="Term",
    )

    references: Annotated[
        List[OntologySourceReference],
        Field(
            description="List of ontology sources used in ISA metadata files.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Source Name",
            },
        ),
    ] = []


class InvestigationPublications(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["publications"],
        section_header="INVESTIGATION PUBLICATIONS",
        section_prefix=None,
    )

    publications: Annotated[
        List[Publication],
        Field(
            description="Investigation publications.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Publication Title",
            },
        ),
    ] = []


class InvestigationContacts(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["people"],
        section_header="INVESTIGATION CONTACTS",
        section_prefix="Person",
    )

    people: Annotated[
        List[Person],
        Field(
            description="Investigation contact list.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Last Name",
            },
        ),
    ] = []


class StudyDesignDescriptors(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["design_types"],
        section_header="STUDY DESIGN DESCRIPTORS",
        section_prefix="Design",
    )

    design_types: Annotated[
        List[ExtendedOntologyAnnotation],
        Field(
            description="Design descriptors "
            "defined in study design descriptors section.",
            json_schema_extra={"auto_fill": True, "header_name": "Type"},
        ),
    ] = []


class StudyPublications(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["publications"],
        section_header="STUDY PUBLICATIONS",
        section_prefix=None,
    )

    publications: Annotated[
        List[Publication],
        Field(
            description="Publications defined in study publications section.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Publication Title",
            },
        ),
    ] = []


class StudyFactors(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["factors"],
        section_header="STUDY FACTORS",
        section_prefix="Factor",
    )

    factors: Annotated[
        List[Factor],
        Field(
            description="Factors defined in study factors section.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Name",
            },
        ),
    ] = []


class StudyAssays(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(exclude=True),
    ] = IsaTabConfig(
        field_order=["assays"],
        section_header="STUDY ASSAYS",
        section_prefix="Assay",
    )

    assays: Annotated[
        List[Assay],
        Field(
            description="List of assays defined in study assay section.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "File Name",
            },
        ),
    ] = []


class StudyProtocols(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["protocols"],
        section_header="STUDY PROTOCOLS",
        section_prefix="Protocol",
    )

    protocols: Annotated[
        List[Protocol],
        Field(
            description="Protocols defined in study protocol section.",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Name",
            },
        ),
    ] = []


class StudyContacts(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=["people"],
        section_header="STUDY CONTACTS",
        section_prefix="Person",
    )

    people: Annotated[
        List[Person],
        Field(
            description="Contacts defined in study contacts section",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "search_header": "Last Name",
            },
        ),
    ] = []


class Funder(IsaAbstractModel):
    funder_name: Annotated[str, Field(description="Funder organization full name.")] = (
        ""
    )
    funder_id: Annotated[str, Field(description="Funder ID.")] = ""
    grant_ids: Annotated[list[str], Field(description="Grant ID.")] = []


class Study(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(
            exclude=True,
        ),
    ] = IsaTabConfig(
        field_order=[
            "identifier",
            "title",
            "description",
            "submission_date",
            "public_release_date",
            "file_name",
        ],
        section_header="STUDY",
        section_prefix="Study",
    )

    identifier: Annotated[
        str,
        Field(
            description="A unique identifier, either a temporary identifier "
            "generated by MetaboLights repository.",
            json_schema_extra={"auto_fill": True, "header_name": "Identifier"},
        ),
    ] = ""
    title: Annotated[
        str,
        Field(
            description="A concise phrase used to encapsulate "
            "the purpose and goal of the study.",
            json_schema_extra={"auto_fill": True, "header_name": "Title"},
        ),
    ] = ""
    description: Annotated[
        str,
        Field(
            description="A textual description of the study "
            "with components such as objective or goals.",
            json_schema_extra={"auto_fill": True, "header_name": "Description"},
        ),
    ] = ""
    submission_date: Annotated[
        str,
        Field(
            description="The date on which the study is submitted to an archive. "
            "String formatted as ISO8601 date YYYY-MM-DD",
            json_schema_extra={"auto_fill": True, "header_name": "Submission Date"},
        ),
    ] = ""
    public_release_date: Annotated[
        str,
        Field(
            description="The date on which the study SHOULD be released publicly. "
            "String formatted as ISO8601 date YYYY-MM-DD",
            json_schema_extra={"auto_fill": True, "header_name": "Public Release Date"},
        ),
    ] = ""
    file_name: Annotated[
        str,
        Field(
            description="Name of the Sample Table file "
            "corresponding the definition of that Study.",
            json_schema_extra={"auto_fill": True, "header_name": "File Name"},
        ),
    ] = ""
    study_category: Annotated[
        str,
        Field(description="Study category", json_schema_extra={"auto_fill": False}),
    ] = ""
    template_version: Annotated[
        str,
        Field(description="Template version", json_schema_extra={"auto_fill": False}),
    ] = ""
    sample_template: Annotated[
        str,
        Field(
            description="Study sample file template",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""
    mhd_accession: Annotated[
        str,
        Field(
            description="MetabolomicsHub accession number of the study",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""
    mhd_model_version: Annotated[
        str,
        Field(
            description="MetabolomicsHub model version of the study",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""
    revision_number: Annotated[
        str,
        Field(
            description="Revision number of the study",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""
    revision_date: Annotated[
        str,
        Field(
            description="Revision date of the study",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""
    revision_comment: Annotated[
        str,
        Field(
            description="Revision comment of the study",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""
    dataset_license: Annotated[
        str,
        Field(
            description="Data license of the study",
            json_schema_extra={"auto_fill": False},
        ),
    ] = ""
    funders: Annotated[
        list[Funder],
        Field(
            description="Study funders.",
            json_schema_extra={"auto_fill": False},
        ),
    ] = []
    characteristic_types: Annotated[
        list[CharacteristicDefinition],
        Field(
            description="Study characteristic types.",
            json_schema_extra={"auto_fill": False},
        ),
    ] = []
    study_design_descriptors: Annotated[
        StudyDesignDescriptors,
        Field(
            description="Content of study design descriptors section.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = StudyDesignDescriptors()
    study_publications: Annotated[
        StudyPublications,
        Field(
            description="Content of study publications section.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = StudyPublications()
    study_factors: Annotated[
        StudyFactors,
        Field(
            description="Content of study factors section.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = StudyFactors()
    study_assays: Annotated[
        StudyAssays,
        Field(
            description="Study assay section of i_Investigation.txt file. "
            "This section contains study assays and comments.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = StudyAssays()
    study_protocols: Annotated[
        StudyProtocols,
        Field(
            description="Content of study protocols section.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = StudyProtocols()
    study_contacts: Annotated[
        StudyContacts,
        Field(
            description="Content of study contacts section.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = StudyContacts()


class Investigation(BaseSection):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(exclude=True),
    ] = IsaTabConfig(
        field_order=[
            "identifier",
            "title",
            "description",
            "submission_date",
            "public_release_date",
        ],
        section_header="INVESTIGATION",
        section_prefix="Investigation",
        specification_version="1.0",
        specification_date="2016-10-28",
    )

    identifier: Annotated[
        str,
        Field(
            description="Investigation identifier.",
            json_schema_extra={"auto_fill": True, "header_name": "Identifier"},
        ),
    ] = ""
    title: Annotated[
        str,
        Field(
            description="A concise name given to the investigation.",
            json_schema_extra={"auto_fill": True, "header_name": "Title"},
        ),
    ] = ""
    description: Annotated[
        str,
        Field(
            description="A textual description of the investigation.",
            json_schema_extra={"auto_fill": True, "header_name": "Description"},
        ),
    ] = ""
    submission_date: Annotated[
        str,
        Field(
            description="The date on which the investigation was reported "
            "to the MetaboLights repository. "
            "String formatted as ISO8601 date YYYY-MM-DD",
            json_schema_extra={"auto_fill": True, "header_name": "Submission Date"},
        ),
    ] = ""
    public_release_date: Annotated[
        str,
        Field(
            description="The date on which the investigation was released publicly. "
            "String formatted as ISO8601 date YYYY-MM-DD",
            json_schema_extra={"auto_fill": True, "header_name": "Public Release Date"},
        ),
    ] = ""

    ontology_source_references: Annotated[
        OntologySourceReferences,
        Field(
            description="Ontology sources used in the i_Investigation.txt file",
            json_schema_extra={
                "auto_fill": True,
                "header_name": "",
                "inherit_prefix": False,
            },
        ),
    ] = OntologySourceReferences()

    investigation_publications: Annotated[
        InvestigationPublications,
        Field(
            description="All publications "
            "prepared to report results of the investigation.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = InvestigationPublications()

    investigation_contacts: Annotated[
        InvestigationContacts,
        Field(
            description="People details of the investigation.",
            json_schema_extra={"auto_fill": True, "header_name": ""},
        ),
    ] = InvestigationContacts()

    studies: Annotated[
        List[Study], Field(description="Studies carried out in the investigation.")
    ] = []

    def sync_study_comments_from_fields(self):
        if not self.studies:
            return

        study_comment_field_map = OrderedDict(
            [
                ("Revision", "revision_number"),
                ("Revision Date", "revision_date"),
                ("Revision Log", "revision_comment"),
                ("License", "dataset_license"),
                ("MHD Accession", "mhd_accession"),
                ("MHD Model Version", "mhd_model_version"),
                ("Study Category", "study_category"),
                ("Template Version", "template_version"),
                ("Sample Template", "sample_template"),
            ]
        )
        for study in self.studies:
            comment_values = OrderedDict()
            comment_names = set()
            for comment_name, field_name in study_comment_field_map.items():
                val = getattr(study, field_name or "")
                if val:
                    comment_values[comment_name] = val
                    comment_names.add(comment_name.lower())
            old_study_comments = study.comments
            study.comments = [
                Comment(name=x, value=[y]) for x, y in comment_values.items()
            ]

            funder_comments = OrderedDict(
                [("Funder", []), ("Funder ROR Id", []), ("Grant Identifier", [])]
            )

            if study.funders:
                for funder in study.funders:
                    if (
                        funder.funder_name
                        or funder.funder_id
                        or ";".join(funder.grant_ids)
                    ):
                        funder_comments["Funder"].append(funder.funder_name)
                        funder_comments["Funder ROR Id"].append(funder.funder_id)
                        funder_comments["Grant Identifier"].append(
                            ";".join(funder.grant_ids)
                        )
            self.add_non_empty_comments(study.comments, funder_comments, comment_names)

            characteristic_comments = OrderedDict(
                [
                    ("Study Characteristics Name", []),
                    ("Study Characteristics Type", []),
                    ("Study Characteristics Type Term Source REF", []),
                    ("Study Characteristics Type Term Accession Number", []),
                    ("Study Characteristics Value Format", []),
                ]
            )
            if study.characteristic_types:
                for characteristic in study.characteristic_types:
                    characteristic_comments["Study Characteristics Name"].append(
                        characteristic.name
                    )
                    characteristic_comments["Study Characteristics Type"].append(
                        characteristic.type
                    )
                    characteristic_comments[
                        "Study Characteristics Type Term Accession Number"
                    ].append(characteristic.term_accession_number)
                    characteristic_comments[
                        "Study Characteristics Type Term Source REF"
                    ].append(characteristic.term_source_ref)
                    characteristic_comments[
                        "Study Characteristics Value Format"
                    ].append(characteristic.value_format)

            self.add_non_empty_comments(
                study.comments,
                characteristic_comments,
                comment_names,
                old_study_comments,
            )

    def sync_study_design_descriptor_comments_from_fields(self):
        if not self.studies:
            return
        for study in self.studies:
            descriptor_comments = OrderedDict(
                [("Study Design Category", []), ("Study Design Source", [])]
            )
            old_comments = study.study_design_descriptors.comments
            comment_names = set()
            if study.study_design_descriptors.design_types:
                study.study_design_descriptors.comments = []
                for characteristic in study.study_design_descriptors.design_types:
                    descriptor_comments["Study Design Category"].append(
                        characteristic.category
                    )
                    descriptor_comments["Study Design Source"].append(
                        characteristic.source
                    )
            self.add_non_empty_comments(
                study.study_design_descriptors.comments,
                descriptor_comments,
                comment_names,
                old_comments,
            )

    def sync_study_factor_comments_from_fields(self):
        if not self.studies:
            return
        for study in self.studies:
            factor_comments = OrderedDict([("Study Factor Value Format", [])])
            old_comments = study.study_factors.comments
            comment_names = set()

            if study.study_factors.factors:
                study.study_factors.comments = []
                comment_names = set([x.lower() for x in factor_comments])
                for characteristic in study.study_factors.factors:
                    factor_comments["Study Factor Value Format"].append(
                        characteristic.value_format or ""
                    )
            self.add_non_empty_comments(
                study.study_factors.comments,
                factor_comments,
                comment_names,
                old_comments,
            )

    def sync_protocol_parameter_comments_from_fields(self):
        if not self.studies:
            return
        for study in self.studies:
            protocol_param_comments = OrderedDict(
                [("Study Protocol Parameters Value Format", [])]
            )
            old_comments = study.study_protocols.comments
            comment_names = set()
            protocols = study.study_protocols.protocols
            if protocols:
                study.study_protocols.comments = []
                comment_names = set([x.lower() for x in protocol_param_comments])
                for protocol in protocols:
                    values = [
                        x.value_format or ""
                        for x in protocol.parameters
                        if x.value_format and x.value_format.strip()
                    ]
                    param_value_formats = ""
                    if values:
                        param_value_formats = ";".join(
                            [x.value_format or "" for x in protocol.parameters]
                        )
                    protocol_param_comments[
                        "Study Protocol Parameters Value Format"
                    ].append(param_value_formats or "")

            self.add_non_empty_comments(
                study.study_protocols.comments,
                protocol_param_comments,
                comment_names,
                old_comments,
            )

    def sync_study_contact_comments_from_fields(self):
        if not self.studies:
            return
        for study in self.studies:
            old_comments = study.study_contacts.comments
            comment_names = set()
            study.study_contacts.comments = []

            for comment_name, field_name in [
                ("Study Person ORCID", "orcid"),
                ("Study Person Additional Email", "additional_emails"),
                ("Study Person Affiliation ROR Id", "affiliation_ror_id"),
            ]:
                contact_comments = OrderedDict([(comment_name, [])])
                if study.study_contacts.people:
                    for person in study.study_contacts.people:
                        val = getattr(person, field_name) or ""
                        if isinstance(val, list):
                            val = ";".join(val)
                        contact_comments[comment_name].append(val)

                self.add_non_empty_comments(
                    study.study_contacts.comments, contact_comments, comment_names
                )

            self.add_non_empty_comments(
                study.study_contacts.comments, {}, comment_names, old_comments
            )

    def sync_assay_comments_from_fields(self):
        if not self.studies:
            return
        for study in self.studies:
            assay_comments = OrderedDict([("Assay Identifier", [])])
            result_file_format_comments = OrderedDict(
                [("Assay Result File Format", [])]
            )
            assay_type_comments = OrderedDict(
                [
                    ("Assay Type", []),
                    ("Assay Type Term Accession Number", []),
                    ("Assay Type Term Source REF", []),
                ]
            )
            assay_omics_comments = OrderedDict(
                [
                    ("Omics Type", []),
                    ("Omics Type Term Accession Number", []),
                    ("Omics Type Term Source REF", []),
                ]
            )
            assay_descriptor_comments = OrderedDict(
                [
                    ("Assay Descriptor", []),
                    ("Assay Descriptor Term Accession Number", []),
                    ("Assay Descriptor Term Source REF", []),
                    ("Assay Descriptor Category", []),
                    ("Assay Descriptor Source", []),
                ]
            )
            old_comments = study.study_assays.comments
            comment_names = set()
            if study.study_assays:
                study.study_assays.comments = []

                for assay in study.study_assays.assays:
                    assay_comments["Assay Identifier"].append(
                        assay.assay_identifier or ""
                    )
                    result_file_format_comments["Assay Result File Format"].append(
                        assay.result_file_format or ""
                    )
                    assay_type_comments["Assay Type"].append(
                        assay.assay_type.term or ""
                    )
                    assay_type_comments["Assay Type Term Accession Number"].append(
                        assay.assay_type.term_accession_number
                        if assay.assay_type
                        else ""
                    )
                    assay_type_comments["Assay Type Term Source REF"].append(
                        assay.assay_type.term_source_ref if assay.assay_type else ""
                    )
                    assay_omics_comments["Omics Type"].append(
                        assay.omics_type.term if assay.omics_type else ""
                    )
                    assay_omics_comments["Omics Type Term Accession Number"].append(
                        assay.omics_type.term_accession_number
                        if assay.omics_type
                        else ""
                    )
                    assay_omics_comments["Omics Type Term Source REF"].append(
                        assay.omics_type.term_source_ref if assay.omics_type else ""
                    )

                    for term, field_name in [
                        ("Assay Descriptor", "term"),
                        (
                            "Assay Descriptor Term Accession Number",
                            "term_accession_number",
                        ),
                        ("Assay Descriptor Term Source REF", "term_source_ref"),
                        ("Assay Descriptor Category", "category"),
                        ("Assay Descriptor Source", "source"),
                    ]:
                        assay_descriptor_comments[term].append(
                            ";".join(
                                [
                                    getattr(x, field_name)
                                    for x in assay.assay_descriptors
                                    if x and x.term
                                ]
                            )
                        )
            self.add_non_empty_comments(
                study.study_assays.comments, assay_comments, comment_names
            )
            self.add_non_empty_comments(
                study.study_assays.comments, assay_type_comments, comment_names
            )
            self.add_non_empty_comments(
                study.study_assays.comments, assay_omics_comments, comment_names
            )
            self.add_non_empty_comments(
                study.study_assays.comments,
                assay_descriptor_comments,
                comment_names,
            )
            self.add_non_empty_comments(
                study.study_assays.comments,
                result_file_format_comments,
                comment_names,
                old_comments,
            )

    def add_non_empty_comments(
        self,
        new_comment_list: list[Comment],
        field_comments_map: OrderedDict[str, list[str]],
        new_comment_names: set[str],
        old_comment_list: None | list[Comment] = None,
    ):
        empty = True
        for v in field_comments_map.values():
            non_empty_vals = [x for x in v if str(x)]
            if non_empty_vals:
                empty = False
                break

        if not empty:
            new_comment_names.update([x.lower() for x in field_comments_map.keys()])

            new_comment_list.extend(
                [Comment(name=x, value=y) for x, y in field_comments_map.items()]
            )
        if old_comment_list:
            for old_comment in old_comment_list:
                if old_comment.name.lower() not in new_comment_names:
                    new_comment_list.append(old_comment)

    def sync_comments_from_fields(self):
        self.sync_study_comments_from_fields()
        self.sync_study_design_descriptor_comments_from_fields()
        self.sync_study_contact_comments_from_fields()
        self.sync_assay_comments_from_fields()
        self.sync_study_factor_comments_from_fields()
        self.sync_protocol_parameter_comments_from_fields()

    def sync_fields_from_comments(self):
        self.sync_study_design_descriptors_from_comments()
        self.sync_contact_fields_from_comments()
        self.sync_assay_fields_from_comments()
        self.sync_study_fields_from_comments()
        self.sync_factors_from_comments()
        self.sync_protocols_from_comments()

    def sync_study_design_descriptors_from_comments(self):
        if not self.studies:
            return
        design_descriptor_comment_field_map = {
            "study design category": "category",
            "study design source": "source",
        }
        for study in self.studies:
            if not study.study_design_descriptors.comments:
                continue
            comments_dict: dict[str, list[str]] = {}
            for comment in study.study_design_descriptors.comments:
                if comment.name:
                    comments_dict[comment.name.lower()] = comment.value

            for comment_name, field_name in design_descriptor_comment_field_map.items():
                val = comments_dict.get(comment_name, [])
                if not val:
                    continue
                for idx, comment_item in enumerate(val):
                    types = study.study_design_descriptors.design_types
                    if len(types) > idx:
                        setattr(types[idx], field_name, comment_item or "")

    def sync_factors_from_comments(self):
        if not self.studies:
            return
        factor_comment_field_map = {"study factor value format": "value_format"}
        for study in self.studies:
            if not study.study_factors.comments:
                continue
            comments_dict: dict[str, list[str]] = {}
            for comment in study.study_factors.comments:
                if comment.name:
                    comments_dict[comment.name.lower()] = comment.value

            for comment_name, field_name in factor_comment_field_map.items():
                val = comments_dict.get(comment_name, [])
                if not val:
                    continue
                for idx, comment_item in enumerate(val):
                    factors = study.study_factors.factors
                    if len(factors) > idx:
                        setattr(factors[idx], field_name, comment_item or "")

    def sync_protocols_from_comments(self):
        if not self.studies:
            return
        protocol_comment_field_map = {
            "study protocol parameters value format": "value_format"
        }
        for study in self.studies:
            if not study.study_protocols.comments:
                continue
            comments_dict: dict[str, list[str]] = {}
            for comment in study.study_protocols.comments:
                if comment.name:
                    comments_dict[comment.name.lower()] = comment.value

            for comment_name, field_name in protocol_comment_field_map.items():
                val = comments_dict.get(comment_name, [])
                if not val:
                    continue
                for idx, comment_item in enumerate(val):
                    protocols = study.study_protocols.protocols
                    if len(protocols) > idx:
                        parameter_comments = comment_item.split(";")
                        for param_ind, param in enumerate(parameter_comments):
                            if len(protocols[idx].parameters) > param_ind:
                                setattr(
                                    protocols[idx].parameters[param_ind],
                                    field_name,
                                    param or "",
                                )

    def sync_contact_fields_from_comments(self):
        if not self.studies:
            return
        for study in self.studies:
            if not study.study_contacts.comments:
                continue
            comments_dict: dict[str, list[str]] = {}
            for comment in study.study_contacts.comments:
                if comment.name:
                    comments_dict[comment.name.lower()] = comment.value

            orcid = comments_dict.get("study person orcid", [])
            additional_emails = comments_dict.get("study person additional email", [])
            affiliation_ror_id = comments_dict.get(
                "study person affiliation ror id", []
            )

            for idx, contact in enumerate(study.study_contacts.people):
                if len(orcid) > idx:
                    contact.orcid = orcid[idx] or ""
                if len(affiliation_ror_id) > idx:
                    contact.affiliation_ror_id = affiliation_ror_id[idx] or ""
                if len(additional_emails) > idx:
                    additional_email = additional_emails[idx] or ""
                    additional_email = additional_email.replace(",", ";")
                    additional_email = additional_email.replace(" ", "")
                    contact.additional_emails = [
                        x for x in additional_email.split(";") if x
                    ]

    def sync_assay_fields_from_comments(self):
        if not self.studies:
            return
        for study in self.studies:
            if not study.study_assays.comments:
                continue
            comments_dict: dict[str, list[str]] = {}
            for comment in study.study_assays.comments:
                if comment.name:
                    comments_dict[comment.name.lower()] = comment.value

            identifier = comments_dict.get("assay identifier", [])
            assay_type = comments_dict.get("assay type", [])
            assay_type_term_accession = comments_dict.get(
                "assay type term accession number", []
            )
            assay_type_term_source = comments_dict.get("assay type term source ref", [])

            omics_type = comments_dict.get("omics type", [])
            omics_type_term_accession = comments_dict.get(
                "omics type term accession number", []
            )
            omics_type_term_source = comments_dict.get("omics type term source ref", [])

            assay_descriptor = comments_dict.get("assay descriptor", [])
            assay_descriptor_term_accession = comments_dict.get(
                "assay descriptor term accession number", []
            )
            assay_descriptor_term_source = comments_dict.get(
                "assay descriptor term source ref", []
            )
            assay_descriptor_category = comments_dict.get(
                "assay descriptor category", []
            )
            assay_descriptor_source = comments_dict.get("assay descriptor source", [])
            result_file_format = comments_dict.get("assay result file format", [])

            for idx, assay in enumerate(study.study_assays.assays):
                if len(identifier) > idx:
                    assay.assay_identifier = identifier[idx].strip()
                if len(result_file_format) > idx:
                    assay.result_file_format = result_file_format[idx].strip()
                if len(assay_type) > idx:
                    assay.assay_type.term = assay_type[idx].strip()
                if len(assay_type_term_source) > idx:
                    assay.assay_type.term_source_ref = assay_type_term_source[
                        idx
                    ].strip()
                if len(assay_type_term_accession) > idx:
                    assay.assay_type.term_accession_number = assay_type_term_accession[
                        idx
                    ].strip()

                if len(omics_type) > idx:
                    assay.omics_type.term = omics_type[idx].strip()
                if len(omics_type_term_source) > idx:
                    assay.omics_type.term_source_ref = omics_type_term_source[
                        idx
                    ].strip()
                if len(omics_type_term_accession) > idx:
                    assay.omics_type.term_accession_number = omics_type_term_accession[
                        idx
                    ].strip()
                desc = assay_descriptor[idx] if len(assay_descriptor) > idx else ""
                desc_term_accession = (
                    assay_descriptor_term_accession[idx]
                    if len(assay_descriptor_term_accession) > idx
                    else ""
                )
                desc_term_source = (
                    assay_descriptor_term_source[idx].strip()
                    if len(assay_descriptor_term_source) > idx
                    else ""
                )
                desc_category = (
                    assay_descriptor_category[idx].strip()
                    if len(assay_descriptor_category) > idx
                    else ""
                )
                desc_source = (
                    assay_descriptor_source[idx].strip()
                    if len(assay_descriptor_source) > idx
                    else ""
                )
                if len(desc.split(";")) > 1:
                    desc_parts = desc.split(";")
                    desc_term_accession_parts = desc_term_accession.split(";")
                    desc_term_source_parts = desc_term_source.split(";")
                    desc_category_parts = desc_category.split(";")
                    desc_source_parts = desc_source.split(";")
                    for sub_index, desc in enumerate(desc_parts):
                        assay.assay_descriptors.append(
                            ExtendedOntologyAnnotation(
                                term=desc,
                                term_accession_number=desc_term_accession_parts[
                                    sub_index
                                ]
                                if len(desc_term_accession_parts) > sub_index
                                else "",
                                term_source_ref=desc_term_source_parts[sub_index]
                                if len(desc_term_source_parts) > sub_index
                                else "",
                                category=desc_category_parts[sub_index]
                                if len(desc_category_parts) > sub_index
                                else "",
                                source=desc_source_parts[sub_index]
                                if len(desc_source_parts) > sub_index
                                else "",
                            )
                        )
                else:
                    assay.assay_descriptors.append(
                        ExtendedOntologyAnnotation(
                            term=desc,
                            term_accession_number=desc_term_accession,
                            term_source_ref=desc_term_source,
                            category=desc_category,
                            source=desc_source,
                        )
                    )

    def sync_study_fields_from_comments(self):
        if not self.studies:
            return
        study_comment_field_map = {
            "revision": "revision_number",
            "revision date": "revision_date",
            "revision log": "revision_comment",
            "license": "dataset_license",
            "mhd accession": "mhd_accession",
            "mhd model version": "mhd_model_version",
            "study category": "study_category",
            "template version": "template_version",
            "sample template": "sample_template",
        }
        for study in self.studies:
            if not study.comments:
                continue
            comments_dict: dict[str, list[str]] = {}
            for comment in study.comments:
                if comment.name:
                    comments_dict[comment.name.lower()] = comment.value

            for comment_name, field_name in study_comment_field_map.items():
                val = comments_dict.get(comment_name, [])
                if val:
                    setattr(study, field_name, val[0] if val else "")

            funders = comments_dict.get("funder", [])
            funder_ids = comments_dict.get("funder ror id", [])
            grant_ids = comments_dict.get("grant identifier", [])
            counts = [len(x) for x in (funders, funder_ids, grant_ids) if x and x[0]]

            if counts and len(counts) > 0:
                study.funders = []
                for idx in range(max(counts)):
                    study.funders.append(
                        Funder(
                            funder_name=funders[idx] if len(funders) > idx else "",
                            funder_id=funder_ids[idx] if len(funder_ids) > idx else "",
                            grant_ids=grant_ids[idx].split(";")
                            if len(grant_ids) > idx
                            else "",
                        )
                    )
            characteristic_names = comments_dict.get("study characteristics name", [])
            characteristic_types = comments_dict.get("study characteristics type", [])
            characteristic_type_accessions = comments_dict.get(
                "study characteristics type term accession number", []
            )
            characteristic_type_sources = comments_dict.get(
                "study characteristics type term source ref", []
            )
            characteristic_value_formats = comments_dict.get(
                "study characteristics value format", []
            )
            counts = [
                len(x)
                for x in (
                    characteristic_names,
                    characteristic_types,
                    characteristic_type_accessions,
                    characteristic_type_sources,
                    characteristic_value_formats,
                )
                if x and x[0]
            ]
            if counts and len(counts) > 0:
                study.characteristic_types = []
                for idx in range(max(counts)):
                    study.characteristic_types.append(
                        CharacteristicDefinition(
                            name=characteristic_names[idx]
                            if len(characteristic_names) > idx
                            else "",
                            type=characteristic_types[idx]
                            if len(characteristic_types) > idx
                            else "",
                            term_accession_number=characteristic_type_accessions[idx]
                            if len(characteristic_type_accessions) > idx
                            else "",
                            term_source_ref=characteristic_type_sources[idx]
                            if len(characteristic_type_sources) > idx
                            else "",
                            value_format=characteristic_value_formats[idx]
                            if len(characteristic_value_formats) > idx
                            else "",
                        )
                    )



import logging
from pathlib import Path

logger = logging.getLogger(__name__)


class MetabolightsSearchUtils:
    @staticmethod
    def get_isa_metadata_files(folder_path: str, recursive: bool = False) -> list[str]:
        metadata_files = []
        patterns = ["[asi]_*.txt", "m_*.tsv"]
        folder = Path(folder_path)
        if not folder.exists() or not folder.is_dir():
            logger.warning("Not valid directory %s", folder_path)
            return metadata_files

        for pattern in patterns:
            if recursive:
                metadata_files.extend(folder.rglob(pattern))
            else:
                metadata_files.extend(list(folder.glob(pattern)))
        return metadata_files



from typing import List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.common import MetabolightsBaseModel
from metabolights_utils.models.parser.enums import ParserMessageType


class ParserMessage(MetabolightsBaseModel):
    type: Annotated[
        ParserMessageType,
        Field(
            description="Message type. Valid values: INFO, WARNING, ERROR, CRITICAL."
        ),
    ] = ParserMessageType.WARNING
    short: Annotated[str, Field(description="Short information about message.")] = ""
    detail: Annotated[str, Field(description="Detailed information about message.")] = (
        ""
    )
    section: Annotated[
        str, Field(description="Section name if it is i_Investigation.txt file")
    ] = ""
    line: Annotated[
        str,
        Field(
            description="File line number. It is valid only for i_Investigation.txt file."
        ),
    ] = ""
    column: Annotated[str, Field(description="ISA file table column name.")] = ""

    def __str__(self) -> str:
        return "\t".join(
            [
                self.type,
                f"section: {self.section}",
                f"line: {self.line}",
                f"column: {self.column}",
                self.short,
                f"detail: {self.detail}",
            ]
        )


class ParserReport(MetabolightsBaseModel):
    messages: Annotated[List[ParserMessage], Field(description="")] = []



from enum import Enum
from typing import Dict, List, Union

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.common import CamelCaseModel


class TsvColumnData(CamelCaseModel):
    header_name: Annotated[str, Field(description="Column header name")] = ""
    values: Annotated[
        Dict[int, str], Field(description="Row indices and cell values of the column")
    ] = {}


class TsvRowData(CamelCaseModel):
    values: Annotated[
        Dict[int, str],
        Field(description="Column indices and cell values of the column"),
    ] = {}


class TsvCellData(CamelCaseModel):
    row_index: Annotated[int, Field(description="Row index of the cell")] = -1
    column_index: Annotated[int, Field(description="Column index of the cell")] = -1
    value: Annotated[str, Field(description="Value of the cell")] = ""


class TsvActionType(str, Enum):
    NO_ACTION = "no-action"
    ADD_ROW = "add-row"
    UPDATE_ROW_DATA = "update-row-data"
    DELETE_ROW = "delete-row"
    MOVE_ROW = "move-row"
    COPY_ROW = "copy-row"
    ADD_COLUMN = "add-column"
    UPDATE_COLUMN_DATA = "update-column-data"
    DELETE_COLUMN = "delete-column"
    MOVE_COLUMN = "move-column"
    COPY_COLUMN = "copy-column"
    UPDATE_COLUMN_HEADER = "update-column-name"
    UPDATE_CELL_DATA = "update-cell-data"


class TsvAction(CamelCaseModel):
    id: Annotated[
        str, Field(description="Id of the action. UUID is used as default")
    ] = ""
    action_type: Annotated[
        TsvActionType,
        Field(description="Tsv action type"),
    ] = TsvActionType.NO_ACTION


class TsvAddRowsAction(TsvAction):
    action_type: TsvActionType = TsvActionType.ADD_ROW
    new_row_indices: Annotated[
        List[int], Field(description="Position (index) of new rows.")
    ] = []
    row_data: Annotated[
        Dict[int, TsvRowData],
        Field(
            description="Row indices and row data."
            + "If there is no row data for the selected row index, an empty row will be added."
            + "If row data has missing column values, missing columns cells will be filled with empty string",
        ),
    ] = {}


class TsvAddColumnsAction(TsvAction):
    action_type: TsvActionType = TsvActionType.ADD_COLUMN
    columns: Annotated[
        Dict[int, TsvColumnData], Field(description="Row indices and column data.")
    ] = {}
    cell_default_values: Annotated[
        Dict[int, str],
        Field(
            description="Column indices and default values of the selected column rows."
        ),
    ] = {}


class TsvDeleteRowsAction(TsvAction):
    action_type: TsvActionType = TsvActionType.DELETE_ROW
    current_row_indices: Annotated[
        List[int], Field(description="Current row indices will be deleted.")
    ] = []


class TsvDeleteColumnsAction(TsvAction):
    action_type: TsvActionType = TsvActionType.DELETE_COLUMN
    current_columns: Annotated[
        Dict[int, str],
        Field(
            description="Current column indices and headers will be deleted. "
            "If column index and header do not match, action will be cancelled.",
        ),
    ] = {}


class TsvMoveRowAction(TsvAction):
    action_type: TsvActionType = TsvActionType.MOVE_ROW
    source_row_index: Annotated[
        int, Field(description="Source row index that will be moved.")
    ] = -1
    new_row_index: Annotated[
        int, Field(description="New row index of the selected row.")
    ] = -1


class TsvMoveColumnAction(TsvAction):
    action_type: TsvActionType = TsvActionType.MOVE_COLUMN
    source_column_index: Annotated[
        int, Field(description="Source column index that will be moved.")
    ] = -1
    source_column_header: Annotated[
        str,
        Field(
            description="Source column header that will be moved to verify the selected column.",
        ),
    ] = ""
    new_column_index: Annotated[
        int, Field(description="New column index of the selected column.")
    ] = -1


class TsvCopyRowAction(TsvAction):
    action_type: str = TsvActionType.COPY_ROW
    source_row_index: Annotated[
        int, Field(description="Source row index that will be copied.")
    ] = -1
    target_row_indices: Annotated[
        List[int], Field(description="Target row indices.")
    ] = []
    selected_column_indices: Annotated[
        List[int],
        Field(
            description="Column indices will be copied. Set empty list to copy all columns of the selected row",
        ),
    ] = []


class TsvCopyColumnAction(TsvAction):
    action_type: TsvActionType = TsvActionType.COPY_COLUMN
    source_column_index: Annotated[int, Field(description="Source colum index")] = -1
    source_column_header: Annotated[
        Union[None, str], Field(description="Source colum header to verify.")
    ] = ""
    target_columns: Annotated[
        Dict[int, str], Field(description="target colum indices and headers.")
    ] = {}
    selected_row_indices: Annotated[
        List[int],
        Field(
            description="Row indices. Set empty list to copy all rows of the selected column",
        ),
    ] = []


class TsvUpdateRowsAction(TsvAction):
    action_type: str = TsvActionType.UPDATE_ROW_DATA
    rows: Annotated[
        Dict[int, TsvRowData],
        Field(
            description="Row indices and updated row data",
        ),
    ] = {}


class TsvUpdateColumnsAction(TsvAction):
    action_type: TsvActionType = TsvActionType.UPDATE_COLUMN_DATA
    columns: Annotated[
        Dict[int, TsvColumnData],
        Field(
            description="Column indices and updated column data",
        ),
    ] = {}


class TsvUpdateCellsAction(TsvAction):
    action_type: TsvActionType = TsvActionType.UPDATE_CELL_DATA
    cells: Annotated[
        List[TsvCellData], Field(description="Updated cell data list")
    ] = []


class TsvUpdateColumnHeaderAction(TsvAction):
    action_type: TsvActionType = TsvActionType.UPDATE_COLUMN_HEADER
    new_headers: Annotated[
        Dict[int, str], Field(description="Column indices and new column headers")
    ] = {}
    current_headers: Annotated[
        Dict[int, str], Field(description="Current column indices and column headers")
    ] = {}


class TsvActionResult(CamelCaseModel):
    action: Annotated[TsvAction, Field(description="Applied action details.")] = (
        TsvAction()
    )
    success: Annotated[bool, Field(description="Result of action.")] = False
    message: Annotated[
        str, Field(description="Error or success message after action.")
    ] = ""


class TsvActionReport(CamelCaseModel):
    results: Annotated[
        List[TsvActionResult],
        Field(description="Ordered results of the applied actions."),
    ] = []
    success: Annotated[bool, Field(description="Result of all applied actions.")] = (
        False
    )
    message: Annotated[str, Field(description="Message after  applied actions.")] = ""
    updated_file_sha256_hash: Annotated[
        str, Field(description="Last SHA256 hash value of the updated file.")
    ] = ""



import pathlib
import uuid

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class MoveColumnTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvMoveColumnAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.MOVE_COLUMN:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvMoveColumnAction = action
        source_column_index = action.source_column_index
        source_column_header = action.source_column_header
        new_column_index = action.new_column_index

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with source_file_path.open("r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")
                new_header_names = []
                if source_column_index < 0 or source_column_index >= len(header_names):
                    result.message = (
                        f"Source column index is not in range. {source_column_index}"
                    )
                    return result
                if new_column_index < 0 or new_column_index >= len(header_names):
                    result.message = (
                        f"New column index is not in range. {new_column_index}"
                    )
                    return result

                moved_header = header_names[source_column_index]
                if moved_header != source_column_header:
                    result.message = (
                        f"Input header name does not match the actual one for the index {source_column_index}."
                        + f"Expected: {source_column_header}, found: {moved_header}"
                    )
                    return result
                new_header_names = [x for x in header_names if x != source_column_index]
                new_header_names.insert(new_column_index, moved_header)

                with target_file_path.open("w", encoding=write_encoding) as target:
                    self.write_row(target, new_header_names)
                    for line in source:
                        row = line.strip("\n").split("\t")
                        moved_data = row[source_column_index]
                        new_row = [x for x in row if x != source_column_index]
                        new_row.insert(new_column_index, moved_data)
                        self.write_row(target, new_row)
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



import os
from typing import Union

import click

from metabolights_utils.provider import definitions
from metabolights_utils.provider.submission_repository import (
    MetabolightsSubmissionRepository,
)

VALID_PARAMETERS = {
    "LC-MS": (
        {"positive", "negative", "alternating"},
        {"hilic", "reverse phase", "direct infusion"},
    ),
    "LC-DAD": (
        {"positive", "negative", "alternating"},
        {"hilic", "reverse phase", "direct infusion"},
    ),
    "GC-MS": ({"positive"}, {}),
    "GCxGC-MS": ({"positive"}, {}),
    "GC-FID": ({"positive"}, {}),
    "DI-MS": ({}, {}),
    "FIA-MS": ({}, {}),
    "CE-MS": ({}, {}),
    "MALDI-MS": ({}, {}),
    "MSImaging": ({}, {}),
    "NMR": ({}, {}),
}


@click.command(no_args_is_help=True, name="create-assay")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_submission_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--rest_api_base_url",
    "-u",
    default=definitions.default_rest_api_url,
    help="MetaboLights study submission API base URL.",
)
@click.option(
    "--local_cache_path",
    "-x",
    default=definitions.default_local_submission_cache_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--credentials_file_path",
    "-c",
    default=definitions.default_local_submission_credentials_file_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--scan_polarity",
    "--polarity",
    help="Scan polarity of the assay.",
)
@click.option(
    "--column_type",
    help="Column type of the assay.",
)
@click.argument("study_id")
@click.argument("assay_technique")
def submission_create_assay(
    study_id: str = "",
    assay_technique: str = "",
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    rest_api_base_url: Union[None, str] = None,
    override_local_files: bool = False,
    credentials_file_path: str = "",
    scan_polarity: str = "",
    column_type: str = "",
):
    """
    Creates a study assay and maf file.

    study_id: MetaboLights study accession number (MTBLSxxxx).
    assay_technique: Valid assay techniques:  NMR, LC-MS, LC-DAD, GC-MS, GCxGC-MS, GC-FID, DI-MS, FIA-MS, CE-MS, MALDI-MS, MSImaging

    Acronyms: Diode array detection (LC-DAD), Tandem MS (GCxGC-MS), Flame ionisation detector (GC-FID),
    Direct infusion (DI-MS), Flow injection analysis (FIA-MS), Capillary electrophoresis (CE-MS),
    Matrix-assisted laser desorption-ionisation imaging mass spectrometry (MALDI-MS),
    Nuclear magnetic resonance (NMR), Mass spec spectrometry (MSImaging)



    scan_polarity (optional): Valid only for LC-MS, LC-DAD, GC-MS, GCxGC-MS, GC-FID. Valid values: positive, negative, alternating


    column_type (optional): Valid only for LC-MS, LC-DAD. Valid values: hilic, reverse phase, direct infusion
    """
    study_id = study_id.upper().strip()
    # assay_technique = assay_technique.upper().strip() if assay_technique else ""
    if assay_technique not in VALID_PARAMETERS:
        click.echo(f"Invalid assay technique: {assay_technique}")
        return
    if scan_polarity and scan_polarity not in VALID_PARAMETERS[assay_technique][0]:
        valid_message = f"{', '.join(VALID_PARAMETERS[assay_technique][0])}"
        if not valid_message:
            valid_message = "There is no scan polarity option for this assay technique. Do not use this parameter option."
        else:
            valid_message = f"Valid scan polarity values are: {valid_message}."
        click.echo(
            f"Invalid scan polarity parameter '{scan_polarity}' for assay technique {assay_technique}. {valid_message}"
        )
        return

    if column_type and column_type not in VALID_PARAMETERS[assay_technique][1]:
        valid_message = f"{', '.join(VALID_PARAMETERS[assay_technique][1])}"
        if not valid_message:
            valid_message = "There is no column type option for this assay technique. Do not use this parameter option."
        else:
            valid_message = f"Valid column types are: {valid_message}."
        click.echo(
            f"Invalid column type  parameter '{column_type}' for assay technique {assay_technique}. {valid_message}"
        )
        return

    client = MetabolightsSubmissionRepository(
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
        rest_api_base_url=rest_api_base_url,
        credentials_file_path=credentials_file_path,
    )
    assay_filename, maf_filename, error = client.create_assay(
        study_id=study_id,
        assay_technique=assay_technique,
        scan_polarity=scan_polarity,
        column_type=column_type,
    )
    if not assay_filename:
        click.echo(error)
        return
    else:
        click.echo(
            f"Created files for study {study_id}: {assay_filename} and {maf_filename}"
        )

    result = client.download_submission_metadata_files(
        study_id=study_id,
        override_local_files=override_local_files,
        delete_unlisted_metadata_files=False,
    )
    study_path = os.path.join(client.local_storage_root_path, study_id)
    click.echo(f"Download submission study {study_id} on {study_path} status:")
    for item in result.actions:
        click.echo(f"  {result.actions[item]}\t{item}")


if __name__ == "__main__":
    submission_create_assay(["MTBLS5397", "DI-MS"])



import os
from typing import Union

import click

from metabolights_utils.provider import definitions
from metabolights_utils.provider.submission_repository import (
    MetabolightsSubmissionRepository,
)


@click.command(no_args_is_help=True, name="delete-assay")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_submission_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--rest_api_base_url",
    "-u",
    default=definitions.default_rest_api_url,
    help="MetaboLights study submission API base URL.",
)
@click.option(
    "--local_cache_path",
    "-x",
    default=definitions.default_local_submission_cache_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--credentials_file_path",
    "-c",
    default=definitions.default_local_submission_credentials_file_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--save_audit_copy",
    "-o",
    is_flag=True,
    default=True,
    help="Save an audit folder before deleting assay file",
)
@click.argument("study_id")
@click.argument("assay_filename")
def submission_delete_assay(
    study_id: str = "",
    assay_filename: str = "",
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    rest_api_base_url: Union[None, str] = None,
    override_local_files: bool = False,
    credentials_file_path: str = "",
    save_audit_copy: bool = True,
):
    """
    Deletes a study assay file and referenced maf file.

    study_id: MetaboLights study accession number (MTBLSxxxx).
    assay_filename: Assay filename




    save_audit_copy (optional): creates an audit folder before deleting study

    """
    study_id = study_id.upper().strip()

    client = MetabolightsSubmissionRepository(
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
        rest_api_base_url=rest_api_base_url,
        credentials_file_path=credentials_file_path,
    )
    success, error = client.delete_assay(
        study_id=study_id,
        assay_filename=assay_filename,
        save_audit_copy=save_audit_copy,
    )
    if not success:
        click.echo(error)
        exit(1)
    else:
        click.echo(
            f"Assay file and related maf file are deleted for study : {study_id}: {assay_filename}"
        )

    result = client.download_submission_metadata_files(
        study_id=study_id,
        override_local_files=override_local_files,
        delete_unlisted_metadata_files=True,
    )
    study_path = os.path.join(client.local_storage_root_path, study_id)
    click.echo(f"Download submission study {study_id} on {study_path} status:")
    for item in result.actions:
        click.echo(f"  {result.actions[item]}\t{item}")


if __name__ == "__main__":
    submission_delete_assay(
        ["MTBLS9776", "a_MTBLS5397_LC-MS_positive_hilic_metabolite_profiling-1.txt"]
    )



import click

from metabolights_utils import __version__
from metabolights_utils.commands.model.model_cli import model_cli
from metabolights_utils.commands.public.public_cli import public_cli
from metabolights_utils.commands.submission.submission_cli import submission_cli


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]}, no_args_is_help=True
)
@click.version_option(__version__)
def cli():
    pass


cli.add_command(public_cli)
cli.add_command(submission_cli)
cli.add_command(model_cli)



import logging
from pathlib import Path
from typing import Any, List, Tuple, Union

from metabolights_utils.isatab import Writer
from metabolights_utils.models.isa.common import IsaTable, IsaTableColumn, IsaTableFile
from metabolights_utils.models.isa.enums import ColumnsStructure
from metabolights_utils.models.metabolights.model import MetabolightsStudyModel

logger = logging.getLogger(__name__)


class IsaFileUtils:
    @staticmethod
    async def save_metabolights_study_model(
        mtbls_model: MetabolightsStudyModel,
        output_dir: str,
        values_in_quotation_mark: bool = False,
        investigation_module_name: Union[None, str] = None,
        sync_comments_from_fields: bool = True,
    ):
        output_dir_path = Path(output_dir)
        if not output_dir_path.exists():
            logger.info("Study save folder %s is created", output_dir)
            output_dir_path.mkdir(parents=True, exist_ok=True)

        i_file_name = mtbls_model.investigation_file_path
        logger.info("Saving investigation file: %s", i_file_name)
        Writer.get_investigation_file_writer().write(
            mtbls_model.investigation,
            f"{output_dir}/{mtbls_model.investigation_file_path}",
            values_in_quotation_mark=values_in_quotation_mark,
            investigation_module_name=investigation_module_name,
            sync_comments_from_fields=sync_comments_from_fields,
        )
        for isa_table_files in (
            mtbls_model.samples,
            mtbls_model.assays,
            mtbls_model.metabolite_assignments,
        ):
            for isa_table_file in isa_table_files.values():
                i_file_name = isa_table_file.file_path
                await IsaFileUtils.save_isa_table(
                    isa_table_file,
                    f"{output_dir}/{isa_table_file.file_path}",
                    values_in_quotation_mark=values_in_quotation_mark,
                )

    @staticmethod
    def sanitise_data(value: Union[None, Any]):
        if isinstance(value, list):
            for idx, val in enumerate(value):
                value[idx] = IsaFileUtils.sanitise_single_value(val)
            return value
        return IsaFileUtils.sanitise_single_value(value)

    @staticmethod
    def sanitise_single_value(value: Union[None, Any]):
        if value is None:
            return ""
        return (
            str(value).replace("\n", " ").replace("\r", " ").replace("\t", " ").strip()
        )

    @staticmethod
    async def save_isa_table(
        isa_table_file: IsaTableFile,
        file_path: str,
        values_in_quotation_mark=False,
    ):
        column_order_map = {}
        column_header_map = {}
        data = isa_table_file.table.data
        for column_model in isa_table_file.table.headers:
            column_order_map[column_model.column_index] = column_model.column_name
            column_header_map[column_model.column_index] = column_model.column_header
        file = Path(file_path)
        basename = file.parent.name
        if not file.parent.exists():
            logger.info("Study save folder %s is created", file.parent)
            file.parent.mkdir(parents=True, exist_ok=True)

        logger.info("Saving file: %s", basename)
        with file.open("w") as f:
            if values_in_quotation_mark:
                header = [
                    f'"{column_header_map[idx]}"'
                    for idx in range(len(column_header_map))
                ]
            else:
                header = [
                    column_header_map[idx].strip('"')
                    for idx in range(len(column_header_map))
                ]
            f.write("\t".join(header) + "\n")

            column_names = [
                column_order_map[idx] for idx in range(len(column_order_map))
            ]
            for row_idx in range(len(data[column_names[0]])):
                row = [data[column_name][row_idx] for column_name in column_names]
                for idx, cell in enumerate(row):
                    cell = IsaFileUtils.sanitise_data(cell) if row[idx] else ""
                    if values_in_quotation_mark:
                        cell = f'"{cell}"'
                    else:
                        cell = cell.strip('"')
                    row[idx] = cell
                f.write("\t".join(row) + "\n")

    @staticmethod
    def add_isa_table_columns(
        mtbls_isa_table: IsaTable,
        header_name: str,
        new_column_index: Union[int, None] = None,
        column_structure: ColumnsStructure = ColumnsStructure.ONTOLOGY_COLUMN,
        default_value: Union[str, None] = None,
    ) -> Tuple[List[str], int]:
        if column_structure not in (
            ColumnsStructure.SINGLE_COLUMN_AND_UNIT_ONTOLOGY,
            ColumnsStructure.ONTOLOGY_COLUMN,
            ColumnsStructure.SINGLE_COLUMN,
        ):
            raise ValueError(
                f"Column structure {column_structure.value} is not valid. "
            )
        table_columns = mtbls_isa_table.columns
        headers = mtbls_isa_table.headers
        selected_index = len(table_columns)
        if (
            new_column_index is not None
            and new_column_index >= 0
            and len(table_columns) >= new_column_index
        ):
            selected_index = new_column_index
        search_headers = [header_name]
        new_column_names = []
        linked_column_names = []
        if column_structure == ColumnsStructure.SINGLE_COLUMN_AND_UNIT_ONTOLOGY:
            linked_column_names = ["Unit", "Term Source REF", "Term Accession Number"]
        elif column_structure == ColumnsStructure.ONTOLOGY_COLUMN:
            linked_column_names = ["Term Source REF", "Term Accession Number"]
        elif column_structure == ColumnsStructure.SINGLE_COLUMN:
            linked_column_names = []

        search_headers.extend(linked_column_names)

        default_value = default_value if default_value else ""
        new_columns = []

        for item in search_headers:
            count = sum([1 for x in headers if x.column_header == item])
            new_column_name = f"{item}.{(count)}" if count > 0 else item
            new_column_names.append(new_column_name)

            if item == header_name:
                categories = header_name.split("[")
                category = ""
                if len(categories) > 1:
                    category = categories[0]
                column_model = IsaTableColumn(
                    column_index=selected_index,
                    column_header=item,
                    column_name=new_column_name,
                    column_category=category,
                    column_prefix=category,
                    additional_columns=linked_column_names,
                    default_value=default_value,
                    column_structure=column_structure,
                )
            else:
                column_model = IsaTableColumn(
                    column_index=selected_index,
                    column_header=item,
                    column_name=new_column_name,
                    column_category="",
                    column_prefix="",
                    column_structure=ColumnsStructure.LINKED_COLUMN,
                )

            new_columns.append(column_model)
            mtbls_isa_table.data[new_column_name] = []
            if mtbls_isa_table.columns and mtbls_isa_table.data:
                first_column_data = mtbls_isa_table.data[mtbls_isa_table.columns[0]]
                mtbls_isa_table.data[new_column_name] = [""] * len(first_column_data)

        if selected_index == len(table_columns):
            updated_header_models = headers + new_columns
            updated_columns = table_columns + new_column_names
        else:
            updated_header_models = (
                headers[:selected_index] + new_columns + headers[selected_index:]
            )

            updated_columns = (
                table_columns[:selected_index]
                + new_column_names
                + table_columns[selected_index:]
            )

        for idx, val in enumerate(updated_columns):
            if idx >= len(table_columns):
                table_columns.append(val)
                headers.append(updated_header_models[idx])
                headers[idx].column_index = idx
            else:
                table_columns[idx] = val
                headers[idx] = updated_header_models[idx]
                headers[idx].column_index = idx
        mtbls_isa_table.column_indices = [x for x in range(len(table_columns))]

        return new_column_names, selected_index



import datetime
import glob
import logging
import os
import re
import subprocess
from typing import Dict, List, Tuple, Union

from metabolights_utils.models.common import GenericMessage, InfoMessage, WarningMessage
from metabolights_utils.models.metabolights.model import (
    StudyFileDescriptor,
    StudyFolderMetadata,
)
from metabolights_utils.provider import definitions
from metabolights_utils.provider.study_provider import AbstractFolderMetadataCollector
from metabolights_utils.provider.utils import is_metadata_filename_pattern

logger = logging.getLogger(__name__)


class LocalFolderMetadataCollector(AbstractFolderMetadataCollector):
    def __init__(self):
        pass

    def visit_folder(
        self,
        directory: str,
        root_path: str,
        metadata: Dict[str, StudyFileDescriptor],
        messages: List[GenericMessage],
        metadata_files_only: bool = False,
        data_files_mapping_folder_name: None | str = None,
    ):
        try:
            dir_relative_path = str(directory).replace(
                f"{str(root_path).rstrip(os.sep)}", ""
            )
            dir_relative_path = dir_relative_path.lstrip("/")
            skip_content = False
            for pattern in definitions.skip_folder_content_patterns:
                if pattern.match(dir_relative_path):
                    skip_content = True
                    break
            if skip_content:
                messages.append(
                    InfoMessage(
                        short=f"{dir_relative_path} directory is in content ignore list. SKIPPED"
                    )
                )
                return
            if data_files_mapping_folder_name and directory == root_path:
                # emulate folder
                parent = os.path.dirname(data_files_mapping_folder_name)
                parent = parent if not parent.startswith(".") else ""
                base = os.path.basename(data_files_mapping_folder_name)
                descriptor = StudyFileDescriptor(
                    base_name=base,
                    file_path=data_files_mapping_folder_name,
                    parent_directory=parent,
                    is_directory=True,
                    is_link=False,
                    created_at=int(datetime.datetime.now().timestamp()),
                    mode="755",
                )
                metadata[data_files_mapping_folder_name] = descriptor

            entries = os.listdir(directory)
            for entry in entries:
                full_path: str = os.path.join(directory, entry)

                if metadata_files_only:
                    if not is_metadata_filename_pattern(entry):
                        continue
                    if not os.path.isfile(full_path):
                        continue

                relative_path = os.path.join(dir_relative_path, entry)
                base_name = os.path.basename(relative_path)
                parent_directory = os.path.dirname(relative_path)
                in_ignore_list = False
                for pattern in definitions.ignore_file_patterns:
                    if pattern.match(relative_path):
                        in_ignore_list = True
                        break
                if in_ignore_list:
                    messages.append(
                        InfoMessage(
                            short=f"{relative_path} is in ignore list. SKIPPED."
                        )
                    )
                    continue

                descriptor = StudyFileDescriptor()

                for tag in definitions.TAG_PATTERNS:
                    for pattern in definitions.TAG_PATTERNS[tag]:
                        if re.match(pattern, base_name, re.IGNORECASE):
                            descriptor.tags.append(tag)
                _, ext = os.path.splitext(relative_path)
                key = relative_path
                key_parent = parent_directory
                if data_files_mapping_folder_name:
                    key = os.path.join(data_files_mapping_folder_name, relative_path)
                    key_parent = os.path.join(
                        data_files_mapping_folder_name, relative_path
                    )
                descriptor.extension = ext
                descriptor.base_name = base_name
                descriptor.parent_directory = key_parent
                descriptor.file_path = key
                descriptor.is_directory = os.path.isdir(full_path)
                descriptor.is_link = os.path.islink(full_path)
                if os.path.exists(full_path):
                    stats = os.stat(full_path)
                    if descriptor.is_directory:
                        descriptor.size_in_bytes = 0
                    else:
                        descriptor.size_in_bytes = stats.st_size
                    descriptor.created_at = int(stats.st_ctime)
                    descriptor.modified_at = int(stats.st_mtime)
                    descriptor.mode = oct(stats.st_mode & 0o777).replace("0o", "")
                metadata[key] = descriptor

                if os.path.isdir(full_path):
                    self.visit_folder(
                        full_path,
                        root_path,
                        metadata=metadata,
                        messages=messages,
                        metadata_files_only=metadata_files_only,
                        data_files_mapping_folder_name=data_files_mapping_folder_name,
                    )

        except PermissionError as ex:
            messages.append(
                WarningMessage(
                    short=f"{directory} directory permission error {str(ex)}"
                )
            )
        except Exception as exc:
            messages.append(
                WarningMessage(short=f"{directory} directory error {str(exc)}")
            )

    def get_folder_metadata(
        self,
        study_path,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
        data_files_path: str = "FILES",
        data_files_mapping_folder_name: None | str = None,
    ) -> Tuple[Union[None, StudyFolderMetadata], List[GenericMessage]]:
        messages: List[GenericMessage] = []
        study_folder_metadata = StudyFolderMetadata()
        metadata: Dict[str, StudyFileDescriptor] = {}
        self.visit_folder(
            study_path,
            study_path,
            metadata=metadata,
            messages=messages,
            metadata_files_only=True,
            data_files_mapping_folder_name=None,
        )
        search_path = data_files_path
        if not os.path.isabs(data_files_path) and not data_files_path.startswith(
            study_path
        ):
            merged_path = os.path.join(study_path, data_files_path)
            if os.path.exists(merged_path):
                search_path = merged_path

        self.visit_folder(
            search_path,
            search_path,
            metadata=metadata,
            messages=messages,
            metadata_files_only=False,
            data_files_mapping_folder_name=data_files_mapping_folder_name,
        )
        study_folder_metadata.folders = {
            x: metadata[x] for x in metadata if metadata[x].is_directory
        }
        study_folder_metadata.files = {
            x: metadata[x] for x in metadata if not metadata[x].is_directory
        }
        data_folder_size = 0
        if calculate_data_folder_size:
            files_folder_path = os.path.join(study_path, "FILES")
            if os.path.exists(files_folder_path):
                size = self.folder_size(files_folder_path)
                data_folder_size = size if size else 0

            study_folder_metadata.folder_size_in_bytes = data_folder_size

        if calculate_metadata_size:
            metadata_size = 0
            metadata_files = glob.glob(f"{study_path}/[asi]_*.txt")
            metadata_files = [x for x in metadata_files]
            maf_files = glob.glob(f"{study_path}/m_*.tsv")
            metadata_files.extend([x for x in maf_files])

            for item in metadata_files:
                stats = os.stat(item)
                metadata_size += stats.st_size

            if study_folder_metadata.folder_size_in_bytes >= 0:
                study_folder_metadata.folder_size_in_bytes += metadata_size
            else:
                study_folder_metadata.folder_size_in_bytes = metadata_size

        total_size = study_folder_metadata.folder_size_in_bytes
        if total_size > -1:
            if total_size / (1024**3) >= 1:
                study_folder_metadata.folder_size_in_str = (
                    str(round(total_size / (1024**3), 2)) + "GB"
                )
            else:
                study_folder_metadata.folder_size_in_str = (
                    str(round(total_size / (1024**2), 2)) + "MB"
                )

        return study_folder_metadata, messages

    def folder_size(self, directory: str) -> Union[int, None]:
        try:
            # Run the 'du' command to get the size of the directory in bytes
            directory = os.path.realpath(directory)
            result = subprocess.run(
                ["du", "-s", directory],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            if result.returncode == 0:
                size_str = result.stdout.split()[0]
                return int(size_str) * 1000
            else:
                logger.error("Error", result.stderr.strip())
                return None
        except FileNotFoundError as e:
            logger.exception(e)
            return None
        except Exception as e:
            logger.exception(e)
            return None


if __name__ == "__main__":
    local = LocalFolderMetadataCollector()
    metadata_path = "/Users/ozgury/work/development/mztabm2mtbls/submission_validation_test/studies/MTBLS263"
    data_files_path = "/Users/ozgury/work/development/mztabm2mtbls/submission_validation_test/studies/MTBLS263/FILES"
    result, messages = local.get_folder_metadata(
        metadata_path,
        data_files_path=data_files_path,
        data_files_mapping_folder_name="FILES",
    )
    pass



import json
import os
from typing import Dict, Union

from metabolights_utils.models.metabolights import metabolights_study, model

default_version = "1.1"


def get_study_model_schema(
    schema_path: Union[str, None] = None, version: None | str = None
) -> Dict:
    if not schema_path:
        if version:
            filename = f"study_model_schema_v{version}.json"
        else:
            filename = f"study_model_schema_v{default_version}.json"
        schema_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), filename)

    with open(schema_path, "r", encoding="utf-8") as sf:
        study_model_jsonschema = json.load(sf)
    return study_model_jsonschema


__all__ = ["metabolights_study", "model", "get_study_model_schema"]



from metabolights_utils.isatab.default import (
    assay_file,
    assignment_file,
    base_isa_file,
    base_isa_table_file,
    factory,
    investigation_file,
    parser,
    sample_file,
    writer,
)

__all__ = [
    "parser",
    "assay_file",
    "assignment_file",
    "base_isa_table_file",
    "base_isa_file",
    "factory",
    "investigation_file",
    "sample_file",
    "writer",
]



import logging
import os
import pathlib
from io import IOBase
from typing import Tuple, Union

logger = logging.getLogger(__name__)


class BaseIsaFile:
    def _get_file_path(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
    ) -> Tuple[Union[str, pathlib.Path, IOBase], str]:
        if not file_buffer_or_path:
            logger.warning("file_buffer_or_path input is not defined.")
            return None
        if isinstance(file_buffer_or_path, IOBase):
            return file_buffer_or_path, os.path.realpath(file_buffer_or_path.name)
        if isinstance(file_buffer_or_path, pathlib.Path):
            return file_buffer_or_path, os.path.realpath(str(file_buffer_or_path))
        elif isinstance(file_buffer_or_path, str):
            return file_buffer_or_path, os.path.realpath(file_buffer_or_path)

    def _get_file_buffer(
        self, file: Union[str, pathlib.Path, IOBase], encoding="utf-8"
    ) -> IOBase:
        if isinstance(file, IOBase):
            file_buffer = file
        elif isinstance(file, pathlib.Path):
            file_buffer = file.open(mode="r", encoding=encoding)
        elif isinstance(file, str):
            file_buffer = open(file=file, mode="r", encoding=encoding)
        else:
            logger.error("Input type is not valid: %s.", type(file))
            raise ValueError("file type is not defined")

        return file_buffer

    def _close_file(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
    ):
        # If we opened file, close it
        if file_buffer_or_path and isinstance(file_buffer_or_path, IOBase):
            file_buffer_or_path.close()



import re
from functools import partial
from io import IOBase
from pathlib import Path
from typing import Callable, List, Tuple, Union

from metabolights_utils.isatab.default.parser.common import (
    SelectedTsvFileContent,
    TsvFileFilterOption,
    read_table_file,
)
from metabolights_utils.models.isa.common import IsaTableColumn, IsaTableFile
from metabolights_utils.models.isa.enums import (
    ColumnsStructure,
    IsaTableAdditionalColumn,
)
from metabolights_utils.models.parser.common import ParserMessage
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.tsv.sort import TsvFileSortOption
from metabolights_utils.utils.hash_utils import MetabolightsHashUtils as HashUtils


def fix_isa_table_file(
    file_path: str,
    messages: List[ParserMessage],
    read_encoding: str,
    write_encoding: str,
    fix_empty_rows: bool = True,
    fix_new_lines_in_cells: bool = True,
    max_iteration_to_fix_new_lines_in_cells: int = 5,
):
    if not fix_new_lines_in_cells and not fix_empty_rows:
        return
    file = Path(file_path)
    basename = file.name
    with file.open("r", encoding=read_encoding) as f:
        file_content = f.read()
        find_empty_lines = None
        find_new_lines_in_cells = None
        if fix_empty_rows:
            find_empty_lines = re.findall(r"[\r\n][\r\n]+", file_content)
            if find_empty_lines:
                messages.append(
                    ParserMessage(
                        type=ParserMessageType.WARNING,
                        short=f"Removed empty lines in {basename}.",
                        detail=f"Removed empty lines in {basename}",
                    )
                )
                file_content = re.sub(r"[\r\n][\r\n]+", r"\n", file_content)
        if fix_new_lines_in_cells:
            results = []

            def replace(x: re.Match):
                groups = x.groups()
                input_val = x.group()
                val = re.sub(r"\s", " ", groups[1].strip())

                result = groups[0] + val + groups[0] + groups[2]
                if result != input_val:
                    results.append((groups[1], val))
                return result

            pattern = r'(")([^"]*)\1(\t|\r|\n|$)'
            new_file_content = re.sub(pattern, replace, file_content)
            if results:
                printed_result = results if len(results) < 20 else results[:20]
                printed_result = [
                    (re.sub("\r\t\n", "{ unexpected char }", x[0]), x[1])
                    for x in printed_result
                ]
                messages.append(
                    ParserMessage(
                        type=ParserMessageType.WARNING,
                        short=f"Removed new line characters and unexpected whitespaces in {basename} file table cells.",
                        detail=f"Removed : {str(printed_result)}",
                    )
                )
                file_content = new_file_content

        if find_empty_lines or find_new_lines_in_cells:
            with file.open("w", encoding=write_encoding) as f:
                f.write(file_content)


def parse_isa_file_content(
    parser: Callable,
    file_path: str,
    messages: List[ParserMessage],
    fix_unicode_exceptions: bool = False,
    remove_empty_rows: bool = False,
    remove_new_lines_in_cells: bool = False,
) -> Tuple[IsaTableFile, List[ParserMessage]]:
    file = Path(file_path)
    try:
        if remove_empty_rows:
            fix_isa_table_file(file_path, messages, "utf-8", "utf-8")

        with file.open("r", encoding="utf-8") as f:
            model = parser(f, messages=messages)
            return model, messages
    except UnicodeDecodeError as ex:
        if fix_unicode_exceptions:
            try:
                if remove_empty_rows:
                    fix_isa_table_file(
                        file_path,
                        messages,
                        "latin-1",
                        "latin-1",
                        remove_empty_rows,
                        remove_new_lines_in_cells,
                        max_iteration_to_fix_new_lines_in_cells=5,
                    )
                with file.open("r", encoding="latin-1") as f:
                    model = parser(f, messages=messages)
                    message = ParserMessage(
                        short="File is read with latin-1 encoding",
                        type=ParserMessageType.WARNING,
                    )
                    message.detail = "File is read with latin-1 encoding"
                    messages.append(message)
                    return model, messages
            except Exception as ex:
                message = ParserMessage(
                    short="File read exception", type=ParserMessageType.CRITICAL
                )
                message.detail = f"File parse error: {str(ex)}"
                messages.append(message)
                return None, messages
        else:
            message = ParserMessage(
                short="File read exception", type=ParserMessageType.CRITICAL
            )
            message.detail = f"File parse error: {str(ex)}"
            messages.append(message)
            return None, messages
    except Exception as ex:
        message = ParserMessage(
            short="File read exception", type=ParserMessageType.CRITICAL
        )
        message.detail = f"File parse error: {str(ex)}"
        messages.append(message)
        return None, messages


def parse_isa_table_sheet_from_fs(
    file_path: str,
    expected_patterns: Union[None, List[List[str]]] = None,
    selected_columns: Union[None, List[str]] = None,
    offset: Union[int, None] = None,
    limit: Union[int, None] = None,
    filter_options: List[TsvFileFilterOption] = None,
    sort_options: List[TsvFileSortOption] = None,
    fix_unicode_exceptions: bool = False,
    remove_empty_rows: bool = False,
    remove_new_lines_in_cells: bool = False,
) -> Tuple[IsaTableFile, List[ParserMessage]]:
    file = Path(file_path)
    basename = file.name
    dirname = file.parent.name
    messages: List[ParserMessage] = []
    if not file_path:
        message = ParserMessage(
            short="File path is not valid", type=ParserMessageType.CRITICAL
        )
        message.detail = "File path is not valid"
        messages.append(message)
        return IsaTableFile(), messages

    if not file.exists() or not file.is_file():
        message = ParserMessage(
            short="File does not exist or it is not a file",
            type=ParserMessageType.CRITICAL,
        )
        message.detail = (
            f"File does not exist or is not a file: {basename} in {dirname}"
        )
        messages.append(message)
        return IsaTableFile(), messages

    file_size = file.stat().st_size
    if file_size == 0:
        message = ParserMessage(short="File is empty", type=ParserMessageType.CRITICAL)
        message.detail = f"File is empty: {basename} in {dirname}"
        messages.append(message)
        return IsaTableFile(), messages

    basename = file.name
    parser = partial(
        get_isa_table_file,
        file_name=basename,
        expected_patterns=expected_patterns,
        selected_columns=selected_columns,
        offset=offset,
        limit=limit,
        filter_options=filter_options,
        sort_options=sort_options,
    )
    read_messages: List[ParserMessage] = []
    table, read_messages = parse_isa_file_content(
        parser,
        file_path=file_path,
        messages=read_messages,
        fix_unicode_exceptions=fix_unicode_exceptions,
        remove_empty_rows=remove_empty_rows,
        remove_new_lines_in_cells=remove_new_lines_in_cells,
    )
    isa_table_file: IsaTableFile = table
    if isa_table_file:
        if file.exists():
            isa_table_file.sha256_hash = HashUtils.sha256sum(file_path)
        if isa_table_file.table:
            messages.extend(read_messages)
            messages = [x for x in messages if x.type != ParserMessageType.INFO]
            return isa_table_file, messages

    messages = [x for x in messages if x.type != ParserMessageType.INFO]
    return isa_table_file, messages


def get_isa_table_file(
    file_buffer_or_path: IOBase,
    file_name: str,
    messages: List[ParserMessage],
    expected_patterns: List[List[str]],
    selected_columns: Union[None, List[str]] = None,
    offset: Union[int, None] = None,
    limit: Union[int, None] = None,
    filter_options: List[TsvFileFilterOption] = None,
    sort_options: List[TsvFileSortOption] = None,
) -> IsaTableFile:
    study_table = IsaTableFile()
    if messages is None:
        messages = []
    if not expected_patterns:
        expected_patterns = []
    file_buffer_or_path.seek(0)
    content: SelectedTsvFileContent = read_table_file(
        file_buffer_or_path,
        messages,
        selected_columns,
        offset,
        limit,
        filter_options,
        sort_options,
    )
    if content is None:
        return study_table
    return update_isa_table_file(
        study_table,
        content,
        file_name,
        messages,
        expected_patterns,
    )


def update_isa_table_file(
    study_table: IsaTableFile,
    content: SelectedTsvFileContent,
    file_name: str,
    messages: List[ParserMessage],
    expected_patterns,
):
    columns = [x.column_name for x in content.columns]
    column_indices = {x.column_name: x.column_index for x in content.columns}
    first_column_name = None
    for column_name in columns:
        if not first_column_name:
            first_column_name = column_name
        cleaned_column_name = get_cleaned_header(column_name)

        if not len(cleaned_column_name.strip()) or len(cleaned_column_name) != len(
            cleaned_column_name.strip()
        ):
            message = ParserMessage(type=ParserMessageType.ERROR)
            column_index = column_indices[column_name]
            message.column = str(column_index)
            if len(cleaned_column_name.strip()) > 0:
                message.short = "Column header starts or ends with space"
                message.detail = (
                    f"'{column_name}' header at column {column_index + 1} "
                    + "ends or starts with space."
                )
            else:
                message.short = "Column header is empty"
                message.detail = (
                    f"'{column_name}' header at column {column_index + 1}  is empty."
                )
            messages.append(message)

    headers = get_headers(
        columns, expected_patterns=expected_patterns, messages=messages
    )
    study_table.table.headers = headers
    study_table.table.total_row_count = content.total_rows
    study_table.table.row_offset = content.offset
    study_table.table.row_count = content.limit
    study_table.table.filtered_total_row_count = content.total_filtered_rows
    study_table.table.selected_column_count = content.selected_column_count
    study_table.table.total_column_count = content.total_columns
    study_table.table.filter_options = content.filter_options
    study_table.table.sort_options = content.sort_options

    filtered_data = {}

    for column in content.columns:
        if not study_table.table.row_indices:
            study_table.table.row_indices = [x for x in column.rows]
        filtered_data[column.column_name] = [column.rows[x] for x in column.rows]
    study_table.table.column_indices = [
        column.column_index for column in content.columns
    ]
    # data = df.to_dict(orient="list")
    study_table.table.data = filtered_data
    study_table.table.columns = columns
    # study_table.table.row_count = len(content.columns[0].rows)
    study_table.file_path = file_name
    return study_table


MULTI_COLUMN_TEMPLATES = {
    ColumnsStructure.ONTOLOGY_COLUMN: {
        "column_names": [
            IsaTableAdditionalColumn.TERM_SOURCE_REF,
            IsaTableAdditionalColumn.TERM_ACCSSION_NUMBER,
        ],
        "searchPatterns": [
            r"(Term Source REF)(\.\d+)?",
            r"(Term Accession Number)(\.\d+)?",
        ],
    },
    ColumnsStructure.SINGLE_COLUMN_AND_UNIT_ONTOLOGY: {
        "column_names": [
            IsaTableAdditionalColumn.UNIT,
            IsaTableAdditionalColumn.TERM_SOURCE_REF,
            IsaTableAdditionalColumn.TERM_ACCSSION_NUMBER,
        ],
        "searchPatterns": [
            r"(Unit)(\.\d+)?",
            r"(Term Source REF)(\.\d+)?",
            r"(Term Accession Number)(\.\d+)?",
        ],
    },
}

additional_column_templates = [
    MULTI_COLUMN_TEMPLATES[ColumnsStructure.SINGLE_COLUMN_AND_UNIT_ONTOLOGY][
        "searchPatterns"
    ],
    MULTI_COLUMN_TEMPLATES[ColumnsStructure.ONTOLOGY_COLUMN]["searchPatterns"],
]

ADDITIONAL_COLUMN_HEADERS = [
    MULTI_COLUMN_TEMPLATES[ColumnsStructure.SINGLE_COLUMN_AND_UNIT_ONTOLOGY][
        "column_names"
    ],
    MULTI_COLUMN_TEMPLATES[ColumnsStructure.ONTOLOGY_COLUMN]["column_names"],
]

multiple_columns_additional_header_patterns = MULTI_COLUMN_TEMPLATES[
    ColumnsStructure.SINGLE_COLUMN_AND_UNIT_ONTOLOGY
]["searchPatterns"]


samples_file_expected_patterns = [
    [r"^(Source Name)$", ""],
    [r"^Characteristics\[(\w[ -~]*)\](\.\d+)?$", "Characteristics"],
    [r"^(Protocol REF)(\.\d+)?$", "Protocol"],
    [r"^(Sample Name)$", ""],
    [r"^Factor Value\[(\w[ -~]*)\](\.\d+)?$", "Factor Value"],
    [r"^Comment\b\[(\w{1}[ -~]*)\](\.\d+)?$", "Comment"],
]

assay_file_expected_patterns: List[List[str]] = [
    [r"^(Extract Name)$", ""],
    [r"^(Protocol REF)(\.\d+)?$", "Protocol"],
    [r"^(Sample Name)$", ""],
    [r"^[ ]*Parameter[ ]+Value[ ]*\[[ ]*(\w[ -~]*)[ ]*\](\.\d+)?$", "Parameter Value"],
    [r"^Comment\b\[(\w{1}[ -~]*)\](\.\d+)?$", "Comment"],
    [r"^(Labeled Extract Name)$", ""],
    [r"^(Label)$", ""],
]


def get_cleaned_header(column_name: str) -> str:
    cleaned_column_name = column_name
    last_dot = column_name.rfind(".")
    if last_dot > 0:
        left = column_name[:last_dot]
        right = column_name[last_dot:].lstrip(".")
        if right and right.isnumeric():
            cleaned_column_name = left
    return cleaned_column_name


def get_headers(
    columns: List[str],
    expected_patterns: List[List[str]],
    messages: List[ParserMessage],
):
    headers = []
    column_index = 0
    columns_count = len(columns)
    while column_index < columns_count:
        column_name = columns[column_index]
        cleaned_column_name = get_cleaned_header(column_name)
        column_structure, defined_additional_column_headers = define_column_structure(
            column_index, columns
        )

        result = None
        pattern_index = -1
        expected_header: bool = False
        message = ParserMessage(type=ParserMessageType.INFO)
        for i in range(len(expected_patterns)):
            result = re.search(expected_patterns[i][0], column_name)
            if result and result.groups():
                pattern_index = i
                expected_header = result and result.groups()
                break

        column = IsaTableColumn(
            column_index=column_index, column_structure=column_structure
        )
        column.column_name = column_name
        column.column_header = cleaned_column_name
        linked_columns = []

        if column_structure == ColumnsStructure.ADDITIONAL_COLUMN:
            column.column_category = "Additional"
            message.type = ParserMessageType.CRITICAL
            message.short = (
                "Additional column header"
                + f"'{cleaned_column_name}'"
                + "is not linked to a previous header."
            )
        elif column_structure == ColumnsStructure.SINGLE_COLUMN:
            update_as_single_column(
                expected_patterns,
                cleaned_column_name,
                pattern_index,
                expected_header,
                message,
                column,
            )
        elif column_structure == ColumnsStructure.INVALID_MULTI_COLUMN:
            update_as_invalid_column(
                expected_patterns,
                cleaned_column_name,
                pattern_index,
                expected_header,
                message,
                column,
            )
        else:
            column.additional_columns = defined_additional_column_headers
            additional_column_index = column_index

            for item in column.additional_columns:
                additional_column: IsaTableAdditionalColumn = item
                additional_column_index = additional_column_index + 1
                linked_column = IsaTableColumn(
                    column_index=additional_column_index,
                    column_structure=ColumnsStructure.LINKED_COLUMN,
                )
                linked_column.column_header = additional_column.value
                linked_column.column_category = "Linked Column"
                linked_column.column_name = columns[additional_column_index]
                linked_columns.append(linked_column)

            update_as_multi_column(
                expected_patterns,
                cleaned_column_name,
                pattern_index,
                expected_header,
                message,
                column,
            )
            column_index = column_index + len(defined_additional_column_headers)

        headers.append(column)
        if linked_columns:
            headers.extend(linked_columns)

        update_message(messages, column_index, message, column)

        column_index = column_index + 1
    return headers


def update_message(
    messages, column_index, message: ParserMessage, column: IsaTableColumn
):
    if not message.detail:
        column_message = " ".join(
            [
                f"Column {(column.column_index + 1):03}: '{column.column_category}' column",
                f"column name: '{column.column_header}'",
            ]
        )
        message.detail = column_message
    if not message.short:
        message.short = f"'{column.column_category}' column: '{column.column_header}'"
    message.column = str(column_index)
    messages.append(message)


def update_as_multi_column(
    expected_patterns,
    cleaned_column_name,
    pattern_index,
    expected_header,
    message: ParserMessage,
    column: IsaTableColumn,
):
    if expected_header:
        column.column_search_pattern = expected_patterns[pattern_index][0]
        column.column_category = expected_patterns[pattern_index][1]
        column.column_prefix = expected_patterns[pattern_index][1]
    else:
        # column.column_category = "Undefined"
        message.type = ParserMessageType.INFO
        message.short = (
            f"Multi column '{cleaned_column_name}' is not in expected column list."
        )


def update_as_invalid_column(
    expected_patterns,
    cleaned_column_name,
    pattern_index,
    expected_header,
    message: ParserMessage,
    column: IsaTableColumn,
):
    if expected_header:
        column.column_search_pattern = expected_patterns[pattern_index][0]
        column.column_prefix = expected_patterns[pattern_index][1]
        column.column_category = expected_patterns[pattern_index][1]
    else:
        column.column_category = "Invalid"
        message.type = ParserMessageType.CRITICAL
        message.short = (
            f"Multi column '{cleaned_column_name}'"
            + " has invalid or unordered additional headers."
        )


def update_as_single_column(
    expected_patterns,
    cleaned_column_name,
    pattern_index,
    expected_header,
    message: ParserMessage,
    column: IsaTableColumn,
):
    if expected_header:
        column.column_search_pattern = expected_patterns[pattern_index][0]
        column.column_category = expected_patterns[pattern_index][1]
        column.column_prefix = expected_patterns[pattern_index][1]
    else:
        column.column_category = "Undefined"
        message.type = ParserMessageType.INFO
        message.short = (
            f"Single column '{cleaned_column_name}' is not in expected column list."
        )


def define_column_structure(column_index, columns):
    column_is_additional = False
    for additional_column_pattern in multiple_columns_additional_header_patterns:
        if re.match(additional_column_pattern, columns[column_index]):
            column_is_additional = True
            break
    if column_is_additional:
        return ColumnsStructure.ADDITIONAL_COLUMN, []
    selected_additional_columns, additional_columns_count = select_additional_columns(
        column_index, columns
    )
    return evalutate_selected_columns(
        selected_additional_columns, additional_columns_count, column_index, columns
    )


def evalutate_selected_columns(
    selected_additional_columns, additional_columns_count, column_index, columns
):
    columns_count = len(columns)
    if selected_additional_columns:
        if additional_columns_count == 2:
            return ColumnsStructure.ONTOLOGY_COLUMN, selected_additional_columns
        elif additional_columns_count == 3:
            return (
                ColumnsStructure.SINGLE_COLUMN_AND_UNIT_ONTOLOGY,
                selected_additional_columns,
            )
    else:
        if additional_columns_count > 0:
            return ColumnsStructure.INVALID_MULTI_COLUMN, []
        else:
            next_column_is_additional = False
            if column_index + 1 < columns_count:
                for (
                    additional_column_pattern
                ) in multiple_columns_additional_header_patterns:
                    if re.match(additional_column_pattern, columns[column_index + 1]):
                        next_column_is_additional = True
                        break
            if next_column_is_additional:
                return ColumnsStructure.INVALID_MULTI_COLUMN, []

    return ColumnsStructure.SINGLE_COLUMN, []


def select_additional_columns(column_index, columns: List[str]):
    for i in range(len(additional_column_templates)):
        template = additional_column_templates[i]
        next_item = column_index + 1
        maximum_match = 0
        matched = True
        internal_index = 0
        for pattern in template:
            if next_item >= len(columns) or not re.match(pattern, columns[next_item]):
                matched = False
                break
            next_item = next_item + 1
            internal_index = internal_index + 1
            if maximum_match < internal_index:
                maximum_match = internal_index

        if matched:
            return ADDITIONAL_COLUMN_HEADERS[i], maximum_match
    return [], maximum_match



from typing import Union

import click
from jsonpath_ng import parse

from metabolights_utils.commands.utils import print_study_model_summary
from metabolights_utils.provider import definitions
from metabolights_utils.provider.submission_repository import (
    MetabolightsSubmissionRepository,
)


@click.command(no_args_is_help=True, name="describe")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_submission_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--rest_api_base_url",
    "-u",
    default=definitions.default_rest_api_url,
    help="MetaboLights study submission API base URL.",
)
@click.option(
    "--use_only_local",
    "-l",
    is_flag=True,
    default=False,
    help="Use only current local directory without connecting FTP server.",
)
@click.option(
    "--override_local_files",
    "-o",
    is_flag=True,
    default=False,
    help="Downloads files and override current local copies. It is valid if there is no use_only_local option",
)
@click.option(
    "--load_folder_index",
    "-i",
    is_flag=True,
    default=False,
    help="Create a folder index to store file descriptors on FTP seerver. "
    "It is valid if there is no use_only_local option.",
)
@click.option(
    "--local_cache_path",
    "-c",
    default=definitions.default_local_submission_cache_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--credentials_file_path",
    "-c",
    default=definitions.default_local_submission_credentials_file_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.argument("study_id")
@click.argument("jsonpath", required=False)
def submission_describe(
    study_id: str = "",
    jsonpath: str = "",
    use_only_local: bool = False,
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    rest_api_base_url: Union[None, str] = None,
    override_local_files: bool = False,
    credentials_file_path: str = "",
    load_folder_index: bool = False,
):
    """
    View summary of user's submitted (private) study. Run jsonpath expression to filter MetaboLights study model.

    study_id: MetaboLights study accession number (MTBLSxxxx).

    jsonpath (optional): jsonpath expression to filter study model. Print summary of study model if not specified.

        Example jsonpath expressions: "$.investigation.studies[0].title", "$.investigation.studies[0].study_protocols.protocols[*].name", "$.assays[*].*.table.columns[*]" - print column names

        Note: jsonpath expressions should be quoted with double quotes.
    """
    study_id = study_id.upper().strip()
    client = MetabolightsSubmissionRepository(
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
        rest_api_base_url=rest_api_base_url,
        credentials_file_path=credentials_file_path,
    )

    model, errors = client.load_study_model(
        study_id=study_id, use_only_local_path=use_only_local
    )

    if not model:
        if not errors:
            click.echo(f"{study_id}: {[x for x in errors]}")
        else:
            click.echo(f"{study_id}: load error.")
        exit(1)

        # error = True

    # if not error:

    if model and model.investigation and model.investigation.studies:
        if jsonpath:
            try:
                json_data = model.model_dump()

                jsonpath_expr = parse(jsonpath)
                match_values = [match.value for match in jsonpath_expr.find(json_data)]
                if match_values:
                    click.echo(
                        click.style(
                            f"{study_id}: '{jsonpath}' search result:", fg="green"
                        )
                    )
                    if len(match_values) == 1:
                        click.echo(match_values[0])
                    else:
                        for match_item in match_values:
                            click.echo(f"{match_item}")
                else:
                    click.echo(
                        click.style(f"{study_id}: '{jsonpath}' no match", fg="red")
                    )
            except Exception as ex:
                click.echo(f"jsonpath '{jsonpath}' expression error: {str(ex)}")
        else:
            click.echo(
                click.style("MetaboLights submission study model summary.", fg="green")
            )
            print_study_model_summary(model, log=click.echo)
    else:
        click.echo(f"{study_id} is not a valid or public study.")


if __name__ == "__main__":
    submission_describe(["MTBLS397", "-o"])



from metabolights_utils.utils import (
    audit_utils,
    filename_utils,
    hash_utils,
    search_utils,
)

__all__ = [
    "audit_utils",
    "filename_utils",
    "hash_utils",
    "search_utils",
]



import sys

import click

from metabolights_utils.commands.public.public_describe import public_describe
from metabolights_utils.commands.public.public_download import public_download
from metabolights_utils.commands.public.public_list import public_list
from metabolights_utils.commands.public.public_remove import public_remove
from metabolights_utils.commands.public.public_search import public_search


@click.group(name="public")
def public_cli():
    """Commands to use MetaboLights public study data and ISA metadata files."""


public_cli.add_command(public_list)
public_cli.add_command(public_describe)
public_cli.add_command(public_download)
public_cli.add_command(public_remove)
public_cli.add_command(public_search)
if __name__ == "__main__":
    if len(sys.argv) == 1:
        public_cli(["--help"])
    else:
        public_cli()



import datetime
import pathlib
import shutil
import uuid
from typing import Dict, List, Union

from metabolights_utils.tsv.actions.add_column import AddColumnsTsvAction
from metabolights_utils.tsv.actions.add_row import AddRowsTsvAction
from metabolights_utils.tsv.actions.base import BaseTsvAction, TsvActionException
from metabolights_utils.tsv.actions.copy_column import CopyColumnTsvAction
from metabolights_utils.tsv.actions.copy_row import CopyRowTsvAction
from metabolights_utils.tsv.actions.delete_column import DeleteColumnsTsvAction
from metabolights_utils.tsv.actions.delete_row import DeleteRowsTsvAction
from metabolights_utils.tsv.actions.move_column import MoveColumnTsvAction
from metabolights_utils.tsv.actions.move_row import MoveRowTsvAction
from metabolights_utils.tsv.actions.update_cell import UpdateCellsTsvAction
from metabolights_utils.tsv.actions.update_column import UpdateColumnsTsvAction
from metabolights_utils.tsv.actions.update_column_header import (
    UpdateColumnHeadersTsvAction,
)
from metabolights_utils.tsv.actions.update_row import UpdateRowsTsvAction
from metabolights_utils.tsv.model import (
    TsvAction,
    TsvActionReport,
    TsvActionResult,
    TsvActionType,
)
from metabolights_utils.utils.hash_utils import MetabolightsHashUtils as HashUtils

TSV_FILE_ACTIONS: Dict[TsvActionType, BaseTsvAction] = {}
TSV_FILE_ACTIONS[TsvActionType.ADD_ROW] = AddRowsTsvAction()
TSV_FILE_ACTIONS[TsvActionType.DELETE_ROW] = DeleteRowsTsvAction()
TSV_FILE_ACTIONS[TsvActionType.UPDATE_ROW_DATA] = UpdateRowsTsvAction()
TSV_FILE_ACTIONS[TsvActionType.COPY_ROW] = CopyRowTsvAction()
TSV_FILE_ACTIONS[TsvActionType.MOVE_ROW] = MoveRowTsvAction()
TSV_FILE_ACTIONS[TsvActionType.ADD_COLUMN] = AddColumnsTsvAction()
TSV_FILE_ACTIONS[TsvActionType.DELETE_COLUMN] = DeleteColumnsTsvAction()
TSV_FILE_ACTIONS[TsvActionType.COPY_COLUMN] = CopyColumnTsvAction()
TSV_FILE_ACTIONS[TsvActionType.UPDATE_COLUMN_DATA] = UpdateColumnsTsvAction()
TSV_FILE_ACTIONS[TsvActionType.MOVE_COLUMN] = MoveColumnTsvAction()
TSV_FILE_ACTIONS[TsvActionType.UPDATE_COLUMN_HEADER] = UpdateColumnHeadersTsvAction()
TSV_FILE_ACTIONS[TsvActionType.UPDATE_CELL_DATA] = UpdateCellsTsvAction()


class TsvFileUpdater:
    def apply_actions(
        self,
        file_path: Union[str, pathlib.Path],
        file_sha256_hash: str,
        actions: List[TsvAction],
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
        temp_path: str = "/tmp/mtb-utils-temp",
    ) -> TsvActionReport:
        report: TsvActionReport = TsvActionReport()
        if not file_path:
            report.message = "Invalid input for file path"
            return report
        if not file_sha256_hash:
            report.message = "Invalid input for file hash value"
            return report

        if not actions:
            report.message = "Invalid input for action list"
            return report

        if isinstance(file_path, pathlib.Path):
            file = file_path
        else:
            file = pathlib.Path(file_path)

        if not file.exists() or not file.is_file():
            report.message = (
                f"File '{str(file)}' does not exist or it is not a regular file."
            )
            return report
        sha256 = HashUtils.sha256sum(file, convert_to_linux_line_ending=True)

        if sha256 != file_sha256_hash:
            report.message = f"SH256 of '{str(file)}' does not match the input value."
            return report

        task_id = str(uuid.uuid4().hex)
        temp_root_path = (
            pathlib.Path(temp_path.rstrip("/"))
            if temp_path
            else pathlib.Path("/tmp/mtb-utils-temp")
        )
        timestamp = str(int(datetime.datetime.now(datetime.timezone.utc).timestamp()))
        temp_folder = temp_root_path / pathlib.Path(
            f"isa_table_actions/{timestamp}/{task_id}"
        )
        temp_folder.mkdir(parents=True, exist_ok=True)
        file_copy_path = temp_folder / pathlib.Path(f"{file.name}_{task_id}")
        temp_source_file_path = temp_folder / pathlib.Path(
            f".{file.name}_{task_id}_temp_1"
        )
        temp_target_file_path = temp_folder / pathlib.Path(
            f".{file.name}_{task_id}_temp_2"
        )

        source_path = file_copy_path
        target_path = temp_target_file_path
        last_file = None
        try:
            for action in actions:
                if action.action_type not in TSV_FILE_ACTIONS:
                    report.message = "Unsupported action: action.action_type."
                    return report
            shutil.move(file, file_copy_path)
            for action in actions:
                helper = TSV_FILE_ACTIONS[action.action_type]
                result: TsvActionResult = helper.apply_action(
                    source_path,
                    target_path,
                    action,
                    read_encoding=read_encoding,
                    write_encoding=write_encoding,
                )
                last_file = target_path
                source_path = last_file
                target_path = (
                    temp_target_file_path
                    if source_path == temp_source_file_path
                    else temp_source_file_path
                )

                report.results.append(result)
                if not result.success:
                    raise TsvActionException(message=result.message)
            new_sha256 = HashUtils.sha256sum(last_file)
            report.updated_file_sha256_hash = new_sha256
            report.success = True
        except TsvActionException as exc:
            report.message = exc.message
        except Exception as exc:
            report.message = str(exc)
        finally:
            if report.success and last_file:
                shutil.move(last_file, file)
            else:
                shutil.move(file_copy_path, file)
            shutil.rmtree(temp_folder)
        return report



import sys

import click

from metabolights_utils.commands.model.model_create import model_create
from metabolights_utils.commands.model.model_explain import model_explain


@click.group(name="model")
def model_cli():
    """Commands to explain MetaboLights study data model."""


model_cli.add_command(model_explain)
model_cli.add_command(model_create)
if __name__ == "__main__":
    if len(sys.argv) == 1:
        model_cli(["--help"])
    else:
        model_cli()



import datetime
import json
import logging
import os
import re
from typing import Dict, List, Tuple, Union

from pydantic import BaseModel

from metabolights_utils.models.common import GenericMessage
from metabolights_utils.models.enums import GenericMessageType
from metabolights_utils.models.metabolights.model import (
    StudyFileDescriptor,
    StudyFolderMetadata,
)
from metabolights_utils.provider import definitions
from metabolights_utils.provider.ftp.default_ftp_client import DefaultFtpClient
from metabolights_utils.provider.study_provider import AbstractFolderMetadataCollector
from metabolights_utils.utils.filename_utils import join_path

logger = logging.getLogger(__name__)


class FolderIndex(BaseModel):
    update_time: Union[None, datetime.datetime, str] = None
    content: Union[None, StudyFolderMetadata] = None


MANAGED_FOLDERS = {"", "FILES", "FILES/RAW_FILES", "FILES/DERIVED_FILES"}


class FtpFolderMetadataCollector(AbstractFolderMetadataCollector):
    def __init__(
        self,
        client: DefaultFtpClient,
        study_id: str,
        folder_index_file_path: str,
        rebuild_folder_index_file: bool = False,
    ):
        if not client or not client.remote_repository_root_directory or not study_id:
            logger.error("Not valid input.")
            raise Exception("Not valid input.")
        self.client = client
        self.study_id = study_id.upper().strip("/")

        self.folder_index_file_path = (
            join_path(folder_index_file_path)
            if folder_index_file_path
            else join_path(
                definitions.default_local_repority_cache_path,
                study_id,
                "mtbls_index.json",
            )
        )
        logger.debug("Folder index path is %s", self.folder_index_file_path)
        self.rebuild_folder_index_file = rebuild_folder_index_file
        self.remote_study_directory = join_path(
            self.client.remote_repository_root_directory, self.study_id
        )
        self.remote_study_directory = self.remote_study_directory.replace(
            "\\", "/"
        ).rstrip("/")

    def visit_folder(
        self,
        directory: str,
        study_path: str,
        metadata: Dict[str, StudyFileDescriptor],
        messages: List[str],
    ):
        try:
            prefix = f"{str(study_path).rstrip('/')}/"
            directory = directory.replace("\\", "/")
            dir_relative_path = (
                str(directory).replace(prefix, "") if study_path != directory else ""
            )

            skip_content = False
            for pattern in definitions.skip_folder_content_patterns:
                if pattern.match(dir_relative_path):
                    skip_content = True
                    break
            if skip_content:
                logger.debug("%s is in ignore list. SKIPPED.", dir_relative_path)
                messages.append(f"{dir_relative_path} is in ignore list. SKIPPED.")
                return
            directory_input = (
                join_path(self.study_id, dir_relative_path)
                if dir_relative_path
                else self.study_id
            )
            result = self.client.list_directory(directory=directory_input)

            required_raw_data_folder_files = [
                x
                for x in result.descriptors
                if not x.is_directory
                and len(x.base_name) > 3
                and x.base_name[:4] == "acqu"
                and "_" not in x.base_name
                and "." not in x.base_name
            ]
            optional_raw_data_folder_files = [
                x
                for x in result.descriptors
                if (
                    (len(x.base_name) > 2 and x.base_name[:3] in ("fid", "ser"))
                    or x.base_name == "pdata"
                )
                and "_" not in x.base_name
                and "." not in x.base_name
            ]
            selected_descriptors = result.descriptors
            if (
                dir_relative_path not in MANAGED_FOLDERS
                and required_raw_data_folder_files
                and optional_raw_data_folder_files
            ):
                logger.info(
                    "%s directory is raw data folder. Only fid*, ser* acqu* files will be added.",
                    dir_relative_path,
                )
                messages.append(
                    f"{dir_relative_path} directory is raw data folder. fid*, ser* acqu* files will be added."
                )
                selected_descriptors = required_raw_data_folder_files.copy()
                selected_descriptors.extend(
                    [x for x in optional_raw_data_folder_files if not x.is_directory]
                )
            for item in selected_descriptors:
                entry = item.base_name
                full_path: str = join_path(directory, entry).replace("\\", "/")
                relative_path = join_path(dir_relative_path, entry).replace("\\", "/")
                base_name = os.path.basename(relative_path)
                parent_directory = os.path.dirname(relative_path).replace("\\", "/")
                in_ignore_list = False
                for pattern in definitions.ignore_file_patterns:
                    if pattern.match(relative_path):
                        in_ignore_list = True
                        break
                if in_ignore_list:
                    logger.debug(
                        "%s directory is in content ignore list. SKIPPED",
                        relative_path,
                    )
                    messages.append(
                        f"{relative_path} directory is in content ignore list. SKIPPED"
                    )
                    continue

                descriptor = StudyFileDescriptor.model_validate(
                    item.model_dump(by_alias=True)
                )

                for tag in definitions.TAG_PATTERNS:
                    for pattern in definitions.TAG_PATTERNS[tag]:
                        if re.match(pattern, base_name, re.IGNORECASE):
                            descriptor.tags.append(tag)
                _, ext = os.path.splitext(relative_path)
                descriptor.extension = ext
                descriptor.parent_directory = parent_directory
                descriptor.file_path = relative_path

                metadata[relative_path] = descriptor
                if item.is_directory:
                    logger.debug("%s is directory, search content.", relative_path)
                    self.visit_folder(
                        full_path, study_path, metadata=metadata, messages=messages
                    )

        except Exception as exc:
            logger.exception("Directory error %s: %s", directory, str(exc))

    def get_folder_metadata(
        self,
        study_path,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
        metadata_files_only: bool = False,
        data_files_mapping_folder_name: None | str = None,
    ) -> Tuple[Union[None, StudyFolderMetadata], List[GenericMessage]]:
        study_folder_metadata = None
        metadata: Dict[str, StudyFileDescriptor] = {}
        messages: List[GenericMessage] = []
        current_file_index = None
        if os.path.exists(self.folder_index_file_path) and os.path.isfile(
            self.folder_index_file_path
        ):
            logger.info("%s file exists, loading...", self.folder_index_file_path)
            try:
                with open(self.folder_index_file_path, "r", encoding="utf-8") as f:
                    data = json.load(f)
                    current_file_index = FolderIndex.model_validate(
                        data, from_attributes=True
                    )
                    study_folder_metadata = current_file_index.content
                    logger.info(
                        "%s file is loaded successfully.", self.folder_index_file_path
                    )
            except json.JSONDecodeError as ex:
                logger.error(
                    "%s json file decode error: %s",
                    self.folder_index_file_path,
                    str(ex),
                )
                msg = f"{self.folder_index_file_path} file decode error. {str(ex)}"
                messages.append(
                    GenericMessage(
                        type=GenericMessageType.WARNING,
                        short=msg,
                        detail=msg,
                    )
                )
            except Exception as ex:
                logger.exception(
                    "%s json file load error: %s",
                    self.folder_index_file_path,
                    str(ex),
                )
                msg = f"{self.folder_index_file_path} file load error. {str(ex)}"
                messages.append(
                    GenericMessage(
                        type=GenericMessageType.WARNING,
                        short=msg,
                        detail=msg,
                    )
                )

        if not study_folder_metadata or self.rebuild_folder_index_file:
            study_folder_metadata = StudyFolderMetadata()
            logger.info("Build study folder metadata index.")
            self.visit_folder(
                self.remote_study_directory,
                self.remote_study_directory,
                metadata=metadata,
                messages=messages,
            )
            study_folder_metadata.folders = {
                x: metadata[x] for x in metadata if metadata[x].is_directory
            }
            study_folder_metadata.files = {
                x: metadata[x] for x in metadata if not metadata[x].is_directory
            }

            now = datetime.datetime.now(datetime.timezone.utc)
            file_index = FolderIndex(update_time=now, content=study_folder_metadata)
            with open(self.folder_index_file_path, "w", encoding="utf-8") as fw:
                fw.write(file_index.model_dump_json(indent=4))
            msg = f"{self.folder_index_file_path} file is updated."
            logger.info(msg)
            messages.append(
                GenericMessage(
                    type=GenericMessageType.INFO,
                    short=msg,
                    detail=msg,
                )
            )
        else:
            msg = f"{self.folder_index_file_path} file is used."
            logger.info(msg)
            messages.append(
                GenericMessage(
                    type=GenericMessageType.INFO,
                    short=msg,
                    detail=msg,
                )
            )

        return (study_folder_metadata, messages)



# from metabolights_utils.common import CamelCaseModel
# from metabolights_utils.isatab import Reader, Writer
# from metabolights_utils.isatab.default.assay_file import DefaultAssayFileReader
# from metabolights_utils.isatab.default.assignment_file import (
#     DefaultAssignmentFileReader,
# )
# from metabolights_utils.isatab.default.factory import (
#     DefaultIsaTabReaderFactory,
#     DefaultIsaTabWriterFactory,
# )
# from metabolights_utils.isatab.default.investigation_file import (
#     DefaultInvestigationFileReader,
#     DefaultInvestigationFileWriter,
# )
# from metabolights_utils.isatab.default.parser.isa_table_parser import (
#     assay_file_expected_patterns,
#     samples_file_expected_patterns,
# )
# from metabolights_utils.isatab.default.sample_file import DefaultSampleFileReader
# from metabolights_utils.isatab.default.writer import DefaultIsaTableFileWriter
# from metabolights_utils.isatab.reader import (
#     InvestigationFileReader,
#     InvestigationFileReaderResult,
#     IsaTableFileReader,
#     IsaTableFileReaderResult,
#     IsaTabReaderFactory,
# )
# from metabolights_utils.isatab.writer import (
#     InvestigationFileWriter,
#     IsaTableFileWriter,
#     IsaTabWriterFactory,
# )
# from metabolights_utils.models.common import GenericMessage, MetabolightsBaseModel
# from metabolights_utils.models.enums import GenericMessageType
# from metabolights_utils.models.isa.assay_file import AssayFile
# from metabolights_utils.models.isa.assignment_file import AssignmentFile
# from metabolights_utils.models.isa.common import (
#     AssayTechnique,
#     Comment,
#     IsaAbstractModel,
#     IsaTable,
#     IsaTableColumn,
#     IsaTableFile,
#     NumericItem,
#     OntologyItem,
#     OrganismAndOrganismPartPair,
# )
# from metabolights_utils.models.isa.enums import (
#     ColumnsStructure,
#     IsaTableAdditionalColumn,
# )
# from metabolights_utils.models.isa.investigation_file import (
#     Assay,
#     BaseSection,
#     Factor,
#     Investigation,
#     InvestigationContacts,
#     InvestigationPublications,
#     OntologyAnnotation,
#     OntologySourceReference,
#     OntologySourceReferences,
#     Person,
#     Protocol,
#     Publication,
#     Study,
#     StudyAssays,
#     StudyContacts,
#     StudyFactors,
#     StudyProtocols,
#     StudyPublications,
#     ValueTypeAnnotation,
# )
# from metabolights_utils.models.metabolights.metabolights_study import (
#     BaseMetabolightsStudyModel,
# )
# from metabolights_utils.models.parser.common import ParserMessage, ParserReport
# from metabolights_utils.models.parser.enums import ParserMessageType
# from metabolights_utils.tsv.actions.add_column import AddColumnsTsvAction
# from metabolights_utils.tsv.actions.add_row import AddRowsTsvAction
# from metabolights_utils.tsv.actions.base import BaseTsvAction, TsvActionException
# from metabolights_utils.tsv.actions.copy_column import CopyColumnTsvAction
# from metabolights_utils.tsv.actions.copy_row import CopyRowTsvAction
# from metabolights_utils.tsv.actions.delete_column import DeleteColumnsTsvAction
# from metabolights_utils.tsv.actions.delete_row import DeleteRowsTsvAction
# from metabolights_utils.tsv.actions.move_column import MoveColumnTsvAction
# from metabolights_utils.tsv.actions.update_cell import UpdateCellsTsvAction
# from metabolights_utils.tsv.actions.update_column import UpdateColumnsTsvAction
# from metabolights_utils.tsv.actions.update_column_header import (
#     UpdateColumnHeadersTsvAction,
# )
# from metabolights_utils.tsv.actions.update_row import UpdateRowsTsvAction
# from metabolights_utils.tsv.filter import (
#     BetweenEqualCustomFilter,
#     ContainsFilter,
#     CustomFilter,
#     EmptyFilter,
#     EndsWithFilter,
#     EnumContainsCustomFilter,
#     Filter,
#     FilterDataType,
#     FilterOperation,
#     FilterRegistry,
#     GreaterEqualFilter,
#     GreaterFilter,
#     LessEqualFilter,
#     LessFilter,
#     RegexFilter,
#     StartsWithFilter,
#     TsvFileFilterOption,
#     TsvFilterException,
#     ValidDatetimeCustomFilter,
#     ValidNumberCustomFilter,
# )
# from metabolights_utils.tsv.model import (
#     TsvAction,
#     TsvActionReport,
#     TsvActionResult,
#     TsvActionType,
#     TsvAddColumnsAction,
#     TsvAddRowsAction,
#     TsvCellData,
#     TsvColumnData,
#     TsvCopyColumnAction,
#     TsvCopyRowAction,
#     TsvDeleteColumnsAction,
#     TsvDeleteRowsAction,
#     TsvMoveColumnAction,
#     TsvMoveRowAction,
#     TsvRowData,
#     TsvUpdateCellsAction,
#     TsvUpdateColumnHeaderAction,
#     TsvUpdateColumnsAction,
#     TsvUpdateRowsAction,
# )
# from metabolights_utils.tsv.sort import (
#     AbstractSorter,
#     CustomSorter,
#     DateTimeSorter,
#     EnumSorter,
#     FloatSorter,
#     IntegerSorter,
#     Sorter,
#     SorterRegistry,
#     SortType,
#     SortValueClassification,
#     StringSorter,
#     TsvFileSortOption,
#     TsvFileSortValueOrder,
#     TsvSortException,
# )
# from metabolights_utils.tsv.tsv_file_updater import TSV_FILE_ACTIONS, TsvFileUpdater
# from metabolights_utils.utils.audit_utils import MetabolightsAuditUtils
# from metabolights_utils.utils.filename_utils import MetabolightsFileNameUtils
# from metabolights_utils.utils.hash_utils import (
#     EMPTY_FILE_HASH,
#     IsaMetadataFolderHash,
#     MetabolightsHashUtils,
# )
# from metabolights_utils.utils.search_utils import MetabolightsSearchUtils
from metabolights_utils import (
    commands,
    common,
    isa_file_utils,
    isatab,
    models,
    provider,
    tsv,
    utils,
)

__version__ = "1.4.11"

__all__ = [
    "commands",
    "isatab",
    "models",
    "provider",
    "tsv",
    "utils",
    "common",
    "isa_file_utils",
    # "Assay",
    # "BaseSection",
    # "Factor",
    # "Investigation",
    # "InvestigationContacts",
    # "InvestigationPublications",
    # "OntologyAnnotation",
    # "OntologySourceReference",
    # "OntologySourceReferences",
    # "Person",
    # "Protocol",
    # "Publication",
    # "Study",
    # "StudyAssays",
    # "StudyContacts",
    # "StudyFactors",
    # "StudyProtocols",
    # "StudyPublications",
    # "ValueTypeAnnotation",
    # "ColumnsStructure",
    # "IsaTableAdditionalColumn",
    # "IsaAbstractModel",
    # "NumericItem",
    # "AssayFile",
    # "AssignmentFile",
    # "GenericMessage",
    # "GenericMessageType",
    # "MetabolightsBaseModel",
    # "CamelCaseModel",
    # "AssayTechnique",
    # "Comment",
    # "IsaTable",
    # "IsaTableColumn",
    # "IsaTableFile",
    # "OntologyItem",
    # "OrganismAndOrganismPartPair",
    # "BaseMetabolightsStudyModel",
    # "ParserMessage",
    # "ParserMessageType",
    # "ParserReport",
    # "Reader",
    # "DefaultAssayFileReader",
    # "InvestigationFileReader",
    # "DefaultAssignmentFileReader",
    # "IsaTableFileReader",
    # "InvestigationFileReaderResult",
    # "IsaTableFileReaderResult",
    # "DefaultIsaTabReaderFactory",
    # "DefaultIsaTabWriterFactory",
    # "DefaultInvestigationFileReader",
    # "DefaultInvestigationFileWriter",
    # "assay_file_expected_patterns",
    # "samples_file_expected_patterns",
    # "DefaultSampleFileReader",
    # "DefaultIsaTableFileWriter",
    # "IsaTabReaderFactory",
    # "Writer",
    # "InvestigationFileWriter",
    # "IsaTableFileWriter",
    # "IsaTabWriterFactory",
    # # TSV
    # "TsvAction",
    # "TsvActionReport",
    # "TsvActionResult",
    # "TsvActionType",
    # "TsvAddColumnsAction",
    # "TsvAddRowsAction",
    # "TsvCellData",
    # "TsvColumnData",
    # "TsvCopyColumnAction",
    # "TsvCopyRowAction",
    # "TsvDeleteColumnsAction",
    # "TsvDeleteRowsAction",
    # "TsvMoveColumnAction",
    # "TsvMoveRowAction",
    # "TsvRowData",
    # "TsvUpdateCellsAction",
    # "TsvUpdateColumnHeaderAction",
    # "TsvUpdateColumnsAction",
    # "TsvUpdateRowsAction",
    # "Filter",
    # "FilterDataType",
    # "FilterOperation",
    # "TsvFileFilterOption",
    # "TsvFilterException",
    # "CustomFilter",
    # "FilterRegistry",
    # "ContainsFilter",
    # "StartsWithFilter",
    # "EndsWithFilter",
    # "GreaterFilter",
    # "GreaterEqualFilter",
    # "LessFilter",
    # "LessEqualFilter",
    # "EmptyFilter",
    # "RegexFilter",
    # "ValidNumberCustomFilter",
    # "ValidDatetimeCustomFilter",
    # "EnumContainsCustomFilter",
    # "BetweenEqualCustomFilter",
    # "AbstractSorter",
    # "CustomSorter",
    # "DateTimeSorter",
    # "EnumSorter",
    # "FloatSorter",
    # "IntegerSorter",
    # "Sorter",
    # "SorterRegistry",
    # "SortType",
    # "SortValueClassification",
    # "StringSorter",
    # "TsvFileSortOption",
    # "TsvFileSortValueOrder",
    # "TsvSortException",
    # "TSV_FILE_ACTIONS",
    # "TsvFileUpdater",
    # "TsvActionException",
    # "BaseTsvAction",
    # "AddColumnsTsvAction",
    # "AddRowsTsvAction",
    # "CopyColumnTsvAction",
    # "CopyRowTsvAction",
    # "DeleteColumnsTsvAction",
    # "DeleteRowsTsvAction",
    # "MoveColumnTsvAction",
    # "UpdateCellsTsvAction",
    # "UpdateColumnsTsvAction",
    # "UpdateColumnHeadersTsvAction",
    # "UpdateRowsTsvAction",
    # # utils
    # "MetabolightsAuditUtils",
    # "MetabolightsSearchUtils",
    # "MetabolightsFileNameUtils",
    # # hash utils
    # "MetabolightsHashUtils",
    # "IsaMetadataFolderHash",
    # "EMPTY_FILE_HASH",
]
